import numpy
from AnyQt import QtWidgets
from silx.gui.plot import Plot2D
from ewoksdata.data import bliss

from ewoksxrpd.tasks.mask import MaskDetection
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.tasks.utils import data_utils
from ewoksxrpd.gui.forms import input_parameters_mask


__all__ = ["OWMaskDetection"]


class OWMaskDetection(OWTriggerWidget, ewokstaskclass=MaskDetection):
    name = "MaskDetection"
    description = "Detect 'bad' detector pixels"
    icon = "icons/widget.png"
    want_main_area = True

    def __init__(self, *args, **kwargs) -> None:
        self._tabs = QtWidgets.QTabWidget()
        super().__init__(*args, **kwargs)

    def _init_forms(self):
        parameter_info = input_parameters_mask(self.get_default_input_values())
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        layout = self._get_main_layout()
        layout.addWidget(self._tabs)
        for name in ("Image 1", "Image 2", "Ratio"):
            self._tabs.addTab(Plot2D(), name)
        super()._init_main_area()

    def _refresh_non_form_input_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_input_widgets()
            self._refresh_mixed_plots()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._refresh_mixed_plots()

    def _refresh_mixed_plots(self):
        if self._tabs.count() == 0:
            return
        inputs = self.get_task_input_values()
        outputs = self.get_task_output_values()
        self._update_image(0, inputs, outputs)
        self._update_image(1, inputs, outputs)
        self._update_ratio(inputs, outputs)

    def _update_image(self, idx, inputs, outputs):
        plot = self._tabs.widget(idx)
        plot.remove(kind="image")

        image = inputs.get(f"image{idx+1}")
        if not data_utils.is_data(image):
            return
        plot.addImage(bliss.get_image(image, retry_timeout=0))

        mask = outputs.get("mask")
        if not data_utils.is_data(mask):
            return
        plot.setSelectionMask(mask)

    def _update_ratio(self, inputs, outputs):
        plot = self._tabs.widget(2)
        plot.remove(kind="image")
        monitor1 = inputs.get("monitor1")
        monitor2 = inputs.get("monitor2")
        image1 = inputs.get("image1")
        image2 = inputs.get("image2")
        if (
            not monitor1
            or not monitor2
            or not data_utils.is_data(image1)
            or not data_utils.is_data(image2)
        ):
            return
        image1 = bliss.get_image(image1, retry_timeout=0)
        image2 = bliss.get_image(image2, retry_timeout=0)
        monitor1 = bliss.get_data(monitor1, retry_timeout=0)
        monitor2 = bliss.get_data(monitor2, retry_timeout=0)
        with numpy.errstate(divide="ignore", invalid="ignore"):
            if monitor2 > monitor1:
                ratio = image2 / image1
            else:
                ratio = image1 / image2
        plot.addImage(ratio)
        mask = outputs.get("mask")
        if not data_utils.is_data(mask):
            return
        plot.setSelectionMask(mask)



import numpy
import pytest
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksxrpd.blissintegratenosave import OWIntegrateBlissScanNoSave


@pytest.mark.parametrize("ndims", [1, 2])
def test_bliss_integrate_no_save_task(ndims, bliss_task_inputs):
    assert_bliss_integrate_no_save(ndims, bliss_task_inputs)


@pytest.mark.parametrize("ndims", [1, 2])
def test_batch_integrate_no_save_widget(ndims, bliss_task_inputs, qtapp):
    assert_bliss_integrate_no_save(ndims, bliss_task_inputs, qtapp=qtapp)


def assert_bliss_integrate_no_save(ndims, bliss_task_inputs, qtapp=None):
    nb_radial = 500
    nb_azim = 100

    if ndims == 2:
        integration_options = {
            "integrator_name": "integrate2d_ng",
            "nbpt_azim": nb_azim,
            "error_model": "poisson",
        }
    else:
        integration_options = {
            "error_model": "azimuthal",
            "integrator_name": "sigma_clip_ng",
            "extra_options": {"max_iter": 3, "thres": 0},
        }

    inputs = {
        **bliss_task_inputs,
        "integration_options": {
            "npt": nb_radial,
            "method": "no_csr_cython",
            **integration_options,
        },
    }
    output_values = execute_task(
        OWIntegrateBlissScanNoSave.ewokstaskclass
        if qtapp is None
        else OWIntegrateBlissScanNoSave,
        inputs=inputs,
    )

    assert output_values["radial_units"] == "2th_deg"
    assert output_values["radial"].shape == (nb_radial,)

    data = output_values["intensity"]
    spectrum0 = data[0]
    for spectrum in data:
        numpy.testing.assert_allclose(spectrum, spectrum0, atol=1)

    if ndims == 2:
        assert output_values["azimuthal"].shape == (nb_azim,)
    else:
        assert output_values["azimuthal"] is None



import os
from numbers import Number
import matplotlib
import numpy
from numpy.typing import ArrayLike

try:
    import matplotlib.pyplot as plt
    from matplotlib.colors import SymLogNorm, LogNorm
except ModuleNotFoundError:
    plt = SymLogNorm = LogNorm = None

import pyFAI
import pyFAI.azimuthalIntegrator

from .calibrate import calculate_geometry
from .utils import data_utils
from .utils import xrpd_utils
from .utils import pyfai_utils
from .data_access import TaskWithDataAccess

__all__ = [
    "DiagnoseCalibrateSingleResults",
    "DiagnoseCalibrateMultiResults",
    "DiagnoseIntegrate1D",
]


class _DiagnoseTask(
    TaskWithDataAccess,
    optional_input_names=["show", "pause", "filename", "figsize", "dpi", "fontsize"],
    output_names=["saved"],
    register=False,
):
    def prepare(self):
        matplotlib.rc("font", size=self.get_input_value("fontsize", 22))

    def show(self):
        if self.inputs.show:
            if self.inputs.pause and numpy.isfinite(self.inputs.pause):
                plt.pause(self.inputs.pause)
            else:
                plt.show()
        if self.inputs.filename:
            path = os.path.dirname(self.inputs.filename)
            if path:
                os.makedirs(path, exist_ok=True)
            plt.gcf().savefig(
                self.inputs.filename, dpi=self.get_input_value("dpi", 150)
            )
            self.outputs.saved = True
        else:
            self.outputs.saved = False
        plt.close()


class _DiagnoseCalibrateResults(
    _DiagnoseTask,
    input_names=["detector", "calibrant"],
    optional_input_names=["detector_config", "scaling"],
    register=False,
):
    def plot_calibration(
        self,
        ax1,
        ax2,
        image: ArrayLike,
        control_pts: dict,
        geometry: dict,
        energy: Number,
        title=None,
    ):
        scaling = self.get_input_value("scaling", "symlog").lower()
        if scaling == "symlog":
            colornorm = SymLogNorm(
                1,
                base=10,
                vmin=numpy.nanmin(image),
                vmax=numpy.nanmax(image),
            )
        elif scaling == "log":
            colornorm = LogNorm(
                vmin=numpy.nanmin(image),
                vmax=numpy.nanmax(image),
            )

        if title:
            title = f"{title}: "
        else:
            title = ""

        self.prepare()
        ax1.imshow(image, origin="lower", cmap="inferno", norm=colornorm)
        ax1.set_title(f"{title}Rings")
        if control_pts:
            ax2.imshow(image, origin="lower", cmap="inferno", norm=colornorm)
            ax2.set_title(f"{title}Control points")

        # Calibrant rings on 2D pattern
        detector = data_utils.data_from_storage(self.inputs.detector)
        detector_config = data_utils.data_from_storage(
            self.get_input_value("detector_config", None)
        )
        detector_object = pyFAI.detectors.detector_factory(
            detector, config=detector_config
        )
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
        wavelength = xrpd_utils.energy_wavelength(energy)
        ai = pyFAI.azimuthalIntegrator.AzimuthalIntegrator(
            detector=detector_object, **geometry, wavelength=wavelength
        )
        calibrant_object.set_wavelength(wavelength)
        tth = calibrant_object.get_2th()
        ttha = ai.twoThetaArray()
        ax1.contour(
            ttha, levels=tth, cmap="autumn", linewidths=1
        )  # linestyles="dashed"

        # Detected points on 2D pattern
        for label, points in control_pts.items():
            ax2.scatter(points["p1"], points["p0"], label=label, marker=".")

    def show(self):
        # Diagnose
        # plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))
        plt.tight_layout()
        super().show()


class DiagnoseCalibrateSingleResults(
    _DiagnoseCalibrateResults,
    input_names=["image", "geometry", "energy"],
    optional_input_names=["rings"],
):
    """Quality of a single-distance pyFAI calibration"""

    def run(self):
        if self.inputs.filename:
            if os.path.exists(self.inputs.filename):
                self.outputs.saved = True
                return
        if plt is None:
            raise RuntimeError("'matplotlib' is not installed")

        self.prepare()
        rings = self.inputs.rings
        figsize = self.get_input_value("figsize", (16, 8))
        if rings:
            _, (ax1, ax2) = plt.subplots(nrows=1, ncols=2, figsize=figsize)
        else:
            rings = dict()
            _, ax1 = plt.subplots(nrows=1, ncols=1, figsize=figsize)
            ax2 = None

        self.plot_calibration(
            ax1,
            ax2,
            self.get_image(self.inputs.image),
            data_utils.data_from_storage(rings),
            data_utils.data_from_storage(self.inputs.geometry),
            self.inputs.energy,
        )
        self.show()


class DiagnoseCalibrateMultiResults(
    _DiagnoseCalibrateResults,
    input_names=[
        "images",
        "positions",
        "parametrization",
        "parameters",
    ],
    optional_input_names=["show", "pause", "rings"],
):
    """Quality of a multi-distance pyFAI calibration"""

    def run(self):
        if self.inputs.filename:
            if os.path.exists(self.inputs.filename):
                self.outputs.saved = True
                return
        if plt is None:
            raise RuntimeError("'matplotlib' is not installed")
        nimages = len(self.inputs.images)
        rings = self.inputs.rings
        figsize = self.get_input_value("figsize", (16, 8))
        figsize = (figsize[0], figsize[1] * nimages)
        if rings:
            rings = {int(k): v for k, v in rings.items()}
            _, axes = plt.subplots(nrows=nimages, ncols=2, figsize=figsize)
        else:
            rings = {i: dict() for i in range(nimages)}
            _, axes = plt.subplots(nrows=nimages, ncols=1, figsize=(10, 10 * nimages))
            axes = [(ax1, None) for ax1 in axes]

        for image, position, ringsi, (ax1, ax2) in zip(
            self.inputs.images, self.inputs.positions, sorted(rings), axes
        ):
            image = self.get_image(image)
            position = self.get_data(position)
            title = f"position={position}"
            parametrization = data_utils.data_from_storage(self.inputs.parametrization)
            geometry, energy = calculate_geometry(
                parametrization, self.inputs.parameters, position
            )
            control_pts = data_utils.data_from_storage(rings[ringsi])
            self.plot_calibration(
                ax1, ax2, image, control_pts, geometry, energy, title=title
            )
        self.show()


class DiagnoseIntegrate1D(
    _DiagnoseTask,
    input_names=[
        "x",
        "y",
        "xunits",
    ],
    optional_input_names=["calibrant", "energy", "yerror", "scaling"],
):
    """Quality of a pyFAI integration"""

    def run(self):
        if self.inputs.filename:
            if os.path.exists(self.inputs.filename):
                self.outputs.saved = True
                return
        if plt is None:
            raise RuntimeError("'matplotlib' is not installed")
        plt.figure(figsize=(16, 8))
        plt.title("Diffractogram")

        scaling = self.get_input_value("scaling", "squareroot").lower()
        if self.missing_inputs.yerror:
            plt.plot(self.inputs.x, self.inputs.y)
        else:
            yerr = numpy.minimum(3 * self.inputs.y, self.inputs.yerror)
            plt.errorbar(self.inputs.x, self.inputs.y, yerr=yerr)

        plt.yscale(scaling)
        uname, uunits = pyfai_utils.parse_units(self.inputs.xunits)
        plt.xlabel(f"{uname} ({uunits})")
        plt.ylabel("Intensity")
        if self.inputs.calibrant:
            assert self.inputs.energy, "'energy' task parameter is missing"
            self.plot_calibrant_lines()
        self.show()

    def plot_calibrant_lines(self):
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant = pyFAI.calibrant.get_calibrant(calibrant)
        wavelength = xrpd_utils.energy_wavelength(self.inputs.energy)
        calibrant.set_wavelength(wavelength)

        xvalues = calibrant.get_peaks(unit=self.inputs.xunits)
        mask = (xvalues >= min(self.inputs.x)) & (xvalues <= max(self.inputs.x))
        xvalues = xvalues[mask]
        if xvalues.size:
            yvalues = numpy.interp(xvalues, self.inputs.x, self.inputs.y)
            labels = xrpd_utils.calibrant_ring_labels(calibrant)
            labels = numpy.array(labels[: mask.size])[mask]
            ymin = min(self.inputs.y)
            for label, x, ymax in zip(labels, xvalues, yvalues):
                plt.plot([x, x], [ymin, ymax])
                # plt.text(x, ymax, label)



from numbers import Number
from typing import List
import numpy
from numpy.typing import ArrayLike

from silx.gui.colors import Colormap
from silx.gui.plot import PlotWidget
from silx.image.marchingsquares import find_contours
import pyFAI
import pyFAI.detectors

from ..tasks.utils.xrpd_utils import energy_wavelength


def plot_image(plot: PlotWidget, image: ArrayLike, **kwargs) -> str:
    return plot.addImage(image, **kwargs)


def plot_theoretical_rings(
    plot: PlotWidget,
    detector: str,
    calibrant: str,
    energy: Number,
    geometry: dict,
    detector_config=None,
    max_rings=None,
    legend=None,
    **kwargs,
) -> List[str]:
    """Plot theoretical and detected Debye rings"""
    detector_object = pyFAI.detectors.detector_factory(detector, config=detector_config)
    mask = detector_object.mask
    wavelength = energy_wavelength(energy)
    ai = pyFAI.azimuthalIntegrator.AzimuthalIntegrator(
        detector=detector_object, **geometry, wavelength=wavelength
    )
    calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
    calibrant_object.set_wavelength(wavelength)

    levels = calibrant_object.get_2th()
    if max_rings:
        if max_rings < 0:
            max_rings = None
    else:
        max_rings = None
    if max_rings:
        levels = levels[:max_rings]
    image = ai.twoThetaArray()
    legends = list()
    if not legend:
        legend = "theory"
    for i, level in enumerate(levels):
        polygons = find_contours(image, level, mask=mask)
        color = None
        for j, polygon in enumerate(polygons):
            x = polygon[:, 1] + 0.5
            y = polygon[:, 0] + 0.5
            s = plot.addCurve(
                x=x,
                y=y,
                legend=f"{legend}-{i}-{j}",
                linestyle="-",
                resetzoom=False,
                color=color,
                **kwargs,
            )
            legends.append(s)
            if j == 0:
                color = plot.getCurve(s).getColor()
    return legends


def plot_detected_rings(
    plot: PlotWidget, rings: dict, legend=None, **kwargs
) -> List[str]:
    legends = list()
    if not legend:
        legend = "detected"
    cm = Colormap(name="jet", normalization="linear", vmin=0, vmax=len(rings))
    for value, (label, points) in enumerate(rings.items()):
        value = numpy.full_like(points["p1"], value * 100)
        legend = plot.addScatter(
            points["p1"],
            points["p0"],
            value,
            legend=f"{legend}_{label}",
            symbol=".",
            colormap=cm,
            **kwargs,
        )
        legends.append(legend)
    return legends



import json

import numpy
import h5py
from silx.io.dictdump import nxtodict
from silx.io.utils import h5py_read_dataset
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksxrpd.nexus import OWSaveNexusPattern1D


def test_save_nexus_task(tmpdir, setup1, bliss_lab6_scan):
    assert_save_nexus(tmpdir, setup1, None, bliss_lab6_scan)


def test_save_nexus_widget(tmpdir, setup1, qtapp, bliss_lab6_scan):
    assert_save_nexus(tmpdir, setup1, qtapp, bliss_lab6_scan)


def assert_save_nexus(tmpdir, setup1, qtapp, bliss_lab6_scan):
    bliss_scan_url = f"{bliss_lab6_scan}::/2.1"
    inputs = {
        "url": str(tmpdir / "result.h5"),
        "x": numpy.linspace(1, 60, 60),
        "y": numpy.random.random(60),
        "xunits": "2th_deg",
        "header": {
            "energy": 10.2,
            "detector": setup1.detector,
            "geometry": setup1.geometry,
        },
        "metadata": {"dummy": {"test": "test"}},
        "bliss_scan_url": bliss_scan_url,
        "retry_timeout": 5,
    }

    execute_task(
        OWSaveNexusPattern1D.ewokstaskclass if qtapp is None else OWSaveNexusPattern1D,
        inputs=inputs,
    )

    with h5py.File(str(tmpdir / "result.h5")) as root:
        expected = {"instrument", "measurement", "integrate", "dummy"}
        nxprocess = root["results/integrate"]
        assert set(root["results"].keys()) == expected
        numpy.testing.assert_array_equal(nxprocess["integrated/2th"], inputs["x"])
        numpy.testing.assert_array_equal(nxprocess["integrated/intensity"], inputs["y"])
        numpy.testing.assert_array_equal(
            root["results/measurement/integrated"], inputs["y"]
        )

        configuration = nxprocess["configuration"]
        expected_keys = {"ewoks_version", "energy", "detector", "geometry"}
        if configuration.attrs["NX_class"] == "NXnote":
            config = json.loads(configuration["data"][()])
        else:
            config = nxtodict(configuration)
            expected_keys.update({"@NX_class", "energy@units"})
        assert set(config) == expected_keys
        numpy.testing.assert_array_equal(config["energy"], inputs["header"]["energy"])
        assert h5py_read_dataset(root["results/dummy/test"]) == "test"



import os
import h5py
from silx.gui.plot import Plot1D
from silx.io.url import DataUrl

from ewoksxrpd.tasks.nexus import SaveNexusPattern1D
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_nexus


__all__ = ["OWSaveNexusPattern1D"]


class OWSaveNexusPattern1D(OWTriggerWidget, ewokstaskclass=SaveNexusPattern1D):
    name = "SaveNexusPattern1D"
    description = "Save a 1D diffraction pattern in Nexus HDF5 format"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_nexus(self.get_default_input_values())
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        self._plot = Plot1D()
        self.mainArea.layout().addWidget(self._plot)
        super()._init_main_area()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._update_plot()

    def _update_plot(self):
        self._plot.remove(kind="curve")

        url = self.get_task_input_value("url")
        if not url:
            return
        url = DataUrl(url)
        if not os.path.exists(url.file_path()):
            return

        with h5py.File(url.file_path(), "r") as parent:
            data_path = url.data_path()
            if not data_path:
                data_path = "/"
            parent = parent[data_path]
            while "default" in parent.attrs:
                parent = parent[parent.attrs["default"]]
            attrs = dict(parent.attrs)
            if {"axes", "signal"} - attrs.keys():
                return

            xname = attrs["axes"][0]
            yname = attrs["signal"]
            x = parent[xname]
            xunits = x.attrs["units"]
            y = parent[yname]

            self._plot.addCurve(
                x[()], y[()], xlabel=f"{xname} ({xunits})", ylabel=yname
            )



from os import PathLike
import pytest
import numpy
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksxrpd.integrate1d import OWIntegrate1D
from orangecontrib.ewoksxrpd.diagnose_integrate1d import OWDiagnoseIntegrate1D
from .xrpd_theory import Measurement, Setup, xPattern, yPattern


@pytest.mark.parametrize("monitorlist", [False, True])
def test_integrate1d_task(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: xPattern,
    ySampleA: yPattern,
    monitorlist: bool,
):
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        monitorlist=monitorlist,
    )


def test_integrate1d_widget(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: xPattern,
    ySampleA: yPattern,
    qtapp,
):
    assert_integrate1d(imageSetup1SampleA, setup1, xSampleA, ySampleA, tmpdir, qtapp)


# from pyFAI.method_registry import IntegrationMethod
# for method in IntegrationMethod._registry:
#    print(f"{method.split}_{method.algo}_{method.impl}")
#
# {split}_{algo}_{impl}{target}
# split: "no", "bbox", "pseudo", "full"
# algo: "histogram", "lut", "csr"
# impl: "python", "cython", "opencl"


def test_sigma_clip_task(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: xPattern,
    ySampleA: yPattern,
):
    integration_options = {
        "error_model": "azimuthal",
        "method": "no_csr_cython",
        "integrator_name": "sigma_clip_ng",
        "extra_options": {"max_iter": 3, "thres": 0},
    }
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        None,
        integration_options=integration_options,
    )


def test_sigma_clip_widget(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: xPattern,
    ySampleA: yPattern,
    qtapp,
):
    integration_options = {
        "error_model": "azimuthal",
        "method": "no_csr_cython",
        "integrator_name": "sigma_clip_ng",
        "extra_options": {"max_iter": 3, "thres": 0},
    }
    assert_integrate1d(
        imageSetup1SampleA,
        setup1,
        xSampleA,
        ySampleA,
        tmpdir,
        qtapp,
        integration_options=integration_options,
    )


def test_integrate1d_reconfig(
    tmpdir: PathLike,
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    imageSetup2SampleA: Measurement,
    setup2: Setup,
    xSampleA: xPattern,
    ySampleA: yPattern,
):
    assert_integrate1d(imageSetup1SampleA, setup1, xSampleA, ySampleA, tmpdir, None)
    assert_integrate1d(imageSetup2SampleA, setup2, xSampleA, ySampleA, tmpdir, None)
    assert_integrate1d(imageSetup1SampleA, setup1, xSampleA, ySampleA, tmpdir, None)
    assert_integrate1d(imageSetup1SampleA, setup1, xSampleA, ySampleA, tmpdir, None)


def assert_integrate1d(
    measurement: Measurement,
    setup: Setup,
    xpattern: xPattern,
    ypattern: yPattern,
    tmpdir: PathLike,
    qtapp,
    integration_options=None,
    monitorlist: bool = False,
):
    integration_options2 = dict(xpattern.integration_options)
    if integration_options:
        integration_options2.update(integration_options)
    inputs = {
        "image": measurement.image,
        "detector": setup.detector,
        "geometry": setup.geometry,
        "energy": setup.energy,
        "integration_options": integration_options2,
    }
    if monitorlist:
        inputs["monitors"] = [measurement.monitor, 1, None, 1]
        inputs["references"] = [ypattern.monitor, 1, 1, None]
    else:
        inputs["monitor"] = measurement.monitor
        inputs["reference"] = ypattern.monitor

    output_values = execute_task(
        OWIntegrate1D.ewokstaskclass if qtapp is None else OWIntegrate1D, inputs=inputs
    )

    assert output_values["xunits"] == xpattern.units
    numpy.testing.assert_allclose(xpattern.x, output_values["x"], rtol=1e-6)
    atol = ypattern.y.max() * 0.01
    numpy.testing.assert_allclose(ypattern.y, output_values["y"], atol=atol)

    # Set show=True to visualize the calibration results
    filename = tmpdir / "diagnose.png"
    inputs = {
        "x": output_values["x"],
        "y": output_values["y"],
        "xunits": output_values["xunits"],
        "show": False,
        "filename": str(filename),
        # "energy": setup.energy,
        # "calibrant": "LaB6"
    }
    execute_task(
        OWDiagnoseIntegrate1D.ewokstaskclass
        if qtapp is None
        else OWDiagnoseIntegrate1D,
        inputs=inputs,
    )
    assert filename.exists()



import os
from typing import Mapping, Union, Optional
from typing_extensions import TypeGuard
from numbers import Number

import numpy
import h5py
from silx.io.url import DataUrl
from blissdata.h5api import dynamic_hdf5
import fabio.edfimage
import fabio.tifimage


def is_data(data) -> TypeGuard[Union[numpy.ndarray, Number, str, list]]:
    if isinstance(data, (numpy.ndarray, Number)):
        return True
    if isinstance(data, (str, list)) and data:
        return True
    return False


def data_from_storage(data, remove_numpy=True):
    if isinstance(data, numpy.ndarray):
        if not remove_numpy:
            return data
        elif data.ndim == 0:
            return data.item()
        else:
            return data.tolist()
    elif isinstance(data, Mapping):
        return {
            k: data_from_storage(v, remove_numpy=remove_numpy)
            for k, v in data.items()
            if not k.startswith("@")
        }
    else:
        return data


def link_bliss_scan(
    outentry: h5py.Group, bliss_scan_url: Union[str, DataUrl], **options
):
    if isinstance(bliss_scan_url, str):
        bliss_scan_url = DataUrl(bliss_scan_url)
    file_path = bliss_scan_url.file_path()
    data_path = bliss_scan_url.data_path()
    out_filename = outentry.file.filename
    ext_filename = os.path.relpath(out_filename, os.path.dirname(file_path))
    if ".." in ext_filename:
        ext_filename = file_path
    with dynamic_hdf5.File(file_path, mode="r", **options) as root:
        inentry = root[data_path]
        # Link to the entire group
        for groupname in ("instrument", "sample"):
            try:
                if groupname in outentry or groupname not in inentry:
                    continue
            except Exception:  # fixed by bliss PR !5435
                continue
            outentry[groupname] = h5py.ExternalLink(
                ext_filename, inentry[groupname].name
            )
        # Link to all sub groups
        for groupname in ("measurement",):
            if groupname not in inentry:
                continue
            igroup = inentry[groupname]
            if groupname in outentry:
                ogroup = outentry[groupname]
            else:
                ogroup = outentry.create_group(groupname)
                ogroup.attrs["NX_class"] = igroup.attrs["NX_class"]
            for name in igroup.keys():
                if name in ogroup:
                    continue
                if name not in ogroup:
                    ogroup[name] = h5py.ExternalLink(ext_filename, igroup[name].name)


def convert_to_3d(data: Union[numpy.ndarray, Number, str, list]):
    data_arr = numpy.array(data)

    if data_arr.ndim >= 3:
        return data_arr

    if data_arr.ndim == 2:
        return data_arr.reshape(1, *data_arr.shape)

    if data_arr.ndim == 1:
        return data_arr.reshape(1, 1, *data_arr.shape)

    return data_arr.reshape(1, 1, 1)


def save_image(
    data: numpy.ndarray,
    save_path: str,
    save_name: str,
    monitor_data: Optional[Number] = None,
    metadata: Optional[dict] = None,
    ext: str = "edf",
) -> str:
    normalized_data = data / monitor_data if monitor_data else data

    header = dict(
        Summed_monitor_counts=str(monitor_data) if monitor_data else "nan",
        **(metadata if metadata else {}),
    )

    ext = ext.lower()
    img_filepath = f"{save_path}/{save_name}.{ext}"
    if ext.lower() == "edf":
        Image = fabio.edfimage.EdfImage
    elif ext == "tiff" or ext == "tif":
        Image = fabio.tifimage.TifImage
    else:
        raise ValueError(f"Unsupported ext {ext}. Only supports EDF and TIFF.")

    img = Image(data=normalized_data, header=header)
    img.write(img_filepath)

    return img_filepath



from dataclasses import dataclass, field
from typing import Any
from numpy.typing import ArrayLike
import numpy
import pyFAI.units
import pyFAI.azimuthalIntegrator
import pyFAI.calibrant


@dataclass(frozen=True)
class xPattern:
    x: ArrayLike = field(repr=False)
    low_limit: float
    high_limit: float
    units: str

    @property
    def integration_options(self) -> dict:
        return {
            "radial_range_min": self.low_limit,
            "radial_range_max": self.high_limit,
            "nbpt_rad": self.x.size,
            "unit": self.units,
        }


@dataclass(frozen=True)
class yPattern:
    y: ArrayLike = field(repr=False)
    monitor: float
    theory: Any


@dataclass(frozen=True)
class Measurement:
    image: ArrayLike = field(repr=False)
    monitor: float


@dataclass(frozen=True)
class Calibration:
    image: ArrayLike = field(repr=False)
    calibrant: str


@dataclass(frozen=True)
class Setup:
    detector: str
    energy: float  # keV
    geometry: dict

    @property
    def wavelength(self) -> float:  # Angstrom
        return pyFAI.units.hc / (self.energy * 1e10)


def calibration(
    name: str, ai: pyFAI.azimuthalIntegrator.AzimuthalIntegrator, setup: Setup
) -> Calibration:
    calibrant = pyFAI.calibrant.get_calibrant(name)
    calibrant.set_wavelength(setup.wavelength)
    # W = FWHM^2 in rad
    FWHM = 0.2  # deg
    W = numpy.radians(FWHM) ** 2
    image = calibrant.fake_calibration_image(ai, Imax=100, W=W)
    return Calibration(image=image, calibrant=name)


def measurement(
    ai: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    xpattern,
    ypattern,
    mult=2,
):
    x = xpattern.x
    y = mult * ypattern.y
    monitor = mult * ypattern.monitor
    image = ai.calcfrom1d(x, y, dim1_unit=xpattern.units, mask=ai.mask)
    return Measurement(image=image, monitor=monitor)



import os
from AnyQt import QtWidgets
from ewoksxrpd.tasks.ascii import SaveAsciiPattern1D
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_ascii


__all__ = ["OWSaveAsciiPattern1D"]


class OWSaveAsciiPattern1D(OWTriggerWidget, ewokstaskclass=SaveAsciiPattern1D):
    name = "SaveAsciiPattern1D"
    description = "Save a 1D diffraction pattern in ASCII format"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_ascii(self.get_default_input_values())
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        layout = self._get_main_layout()
        self._textedit = QtWidgets.QTextEdit()
        self._textedit.setReadOnly(True)
        layout.addWidget(self._textedit)
        super()._init_main_area()
        self._refresh_non_form_output_widgets()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._update_output_file()

    def _update_output_file(self):
        inputs = self.get_task_input_values()
        filename = inputs.get("filename")
        if not filename or not os.path.isfile(filename):
            self._textedit.clear()
            return

        with open(filename) as f:
            lines = list(f)

        self._textedit.setPlainText("".join(lines))



import os
import numpy
from ewokscore import Task

from .utils.data_utils import is_data
from .utils.pyfai_utils import integration_info_as_text

__all__ = ["SaveAsciiPattern1D"]


class SaveAsciiPattern1D(
    Task,
    input_names=["filename", "x", "y", "xunits"],
    optional_input_names=["header", "yerror", "metadata"],
    output_names=["saved"],
):
    """Save single diffractogram in ASCII format"""

    def run(self):
        if is_data(self.inputs.yerror):
            data = [self.inputs.x, self.inputs.y, self.inputs.yerror]
        else:
            data = [self.inputs.x, self.inputs.y]
        data = numpy.stack(data, axis=1)

        header = self.get_input_value("header", dict())
        metadata = self.get_input_value("metadata", dict())
        lines = integration_info_as_text(header, xunits=self.inputs.xunits, **metadata)

        dirname = os.path.dirname(self.inputs.filename)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        numpy.savetxt(self.inputs.filename, data, header="\n".join(lines))
        self.outputs.saved = True



from typing import Optional
from ewokscore import Task
from blissdata.h5api import dynamic_hdf5
import h5py

from ewoksxrpd.tasks.data_access import TaskWithDataAccess

from .utils.sum_utils import ImageSum, generate_range, save_sum

SUM_TYPES = ["per_scan", "all_scans", "both"]


class SumBlissScanImages(
    TaskWithDataAccess,
    input_names=["filename", "scan", "detector_name", "output_filename"],
    optional_input_names=[
        "monitor_name",
        "subscan",
        "scan_memory_url",
        "output_process",
        "skip_dark",
    ],
    output_names=["output_uri"],
):
    """Sum images of a single camera of a single Bliss scan"""

    def run(self):
        filename: str = self.inputs.filename
        scan: int = self.inputs.scan
        subscan = self.get_input_value("subscan", 1)
        detector_name: str = self.inputs.detector_name
        output_filename: str = self.inputs.output_filename
        output_url = f"silx://{output_filename}?path=/{scan}.{subscan}"
        monitor_name: Optional[str] = self.get_input_value("monitor_name", None)
        output_process: str = self.get_input_value("output_process", "sum")
        skip_dark: bool = self.get_input_value("skip_dark", False)

        if monitor_name:
            counter_names = [monitor_name]
        else:
            counter_names = None

        if self.inputs.scan_memory_url:
            data_iterator = self.iter_bliss_data_from_memory(
                self.inputs.scan_memory_url,
                lima_names=[detector_name],
                counter_names=counter_names,
            )

            # Dark images are even indices when in memory
            def is_dark(image_number):
                return image_number % 2 == 0

        else:
            data_iterator = self.iter_bliss_data(
                filename,
                scan,
                lima_names=[detector_name],
                counter_names=counter_names,
                subscan=subscan,
            )
            with dynamic_hdf5.File(
                filename, lima_names=[detector_name], **self.get_retry_options()
            ) as root:
                dataset_shape = root[
                    f"{scan}.{subscan}/instrument/{detector_name}/data"
                ].shape
                nb_points_in_scan = dataset_shape[0]

            # Dark images are put at the end once the scan is saved
            def is_dark(image_number):
                return image_number >= nb_points_in_scan // 2

        with self.open_h5item(output_url, mode="a", create=True) as output:
            sum_process = output.create_group(output_process)
            sum_process.attrs["NX_class"] = "NXprocess"
            sum_process.attrs["default"] = "results"
            nxdata = sum_process.create_group("results")
            nxdata.attrs["NX_class"] = "NXdata"

            scan_sum = None
            summed_indices = []
            for image_number, ptdata in enumerate(data_iterator):
                if skip_dark and is_dark(image_number):
                    continue

                image = ptdata[detector_name]
                if scan_sum is None:
                    scan_sum = ImageSum(image.shape)

                if monitor_name is not None:
                    monitor = ptdata[monitor_name]
                else:
                    monitor = None

                scan_sum.add_to_sum(image, monitor)
                summed_indices.append(image_number)

            if scan_sum:
                dset = save_sum(
                    nxdata,
                    name=f"Scan{scan}-Images{summed_indices[0]}-{summed_indices[-1]}",
                    image_sum=scan_sum,
                )

                self.outputs.output_uri = f"{nxdata.file.filename}::{dset.name}"


class SumImages(
    Task,
    input_names=["filename", "detector_name", "output_filename"],
    optional_input_names=[
        "start_scan",
        "end_scan",
        "start_image",
        "end_image",
        "block_size",
        "monitor_name",
        "output_entry",
        "output_process",
        "sum_type",
    ],
    output_names=["output_uris"],
):
    """Sum images of a single camera from a Bliss scan file

    For each scan, images are added in blocks of `block_size` images (one block with all images by default).

    The result contains:
        * the block sums when sum_type=per_scan or sum_type=both
        * the sum of the block sums when sum_type=all_scans or sum_type=both
    """

    def run(self):
        filename: str = self.inputs.filename
        detector_name: str = self.inputs.detector_name
        output_filename: str = self.inputs.output_filename
        start_scan: int = self.get_input_value("start_scan", 1)
        end_scan: Optional[int] = self.get_input_value("end_scan", None)
        start_image: int = self.get_input_value("start_image", 0)
        end_image: Optional[int] = self.get_input_value("end_image", None)
        block_size: Optional[int] = self.get_input_value("block_size", None)
        monitor_name: Optional[str] = self.get_input_value("monitor_name", None)
        output_entry: str = self.get_input_value("output_entry", "processing")
        output_process: str = self.get_input_value("output_process", "sum")
        sum_type: str = self.get_input_value("sum_type", "per_scan")

        if sum_type not in SUM_TYPES:
            raise TypeError(
                f"sum_type must be one of the following values: {SUM_TYPES}. Got {sum_type} instead."
            )
        save_scan_sums = sum_type == "per_scan" or sum_type == "both"
        save_full_sum = sum_type == "all_scans" or sum_type == "both"

        with h5py.File(filename, "r") as h5file:
            nscans = len(h5file)
            scan_range = generate_range(start_scan, end_scan, nscans + 1)

            first_scan_name = list(h5file.keys())[0]
            nimages, *detector_shape = h5file[
                f"{first_scan_name}/measurement/{detector_name}"
            ].shape
            image_range = list(generate_range(start_image, end_image, nimages))

            scan_sum = ImageSum(detector_shape)
            full_sum = ImageSum(detector_shape)

            with h5py.File(output_filename, "a") as output:
                entry = output.require_group(output_entry)
                entry.attrs["NX_class"] = "NXentry"
                entry.attrs["default"] = output_process
                sum_process = entry.create_group(output_process)
                sum_process.attrs["NX_class"] = "NXprocess"
                sum_process.attrs["default"] = "results"
                results = sum_process.create_group("results")
                results.attrs["NX_class"] = "NXdata"

                output_uris = []
                for scan_number in scan_range:
                    scan_images = h5file[f"{scan_number}.1/measurement/{detector_name}"]
                    monitor_data = (
                        h5file[f"{scan_number}.1/measurement/{monitor_name}"]
                        if monitor_name
                        else None
                    )
                    assert isinstance(scan_images, h5py.Dataset)
                    if monitor_data is not None:
                        assert isinstance(monitor_data, h5py.Dataset)

                    for image_number in image_range:
                        image = scan_images[image_number]
                        monitor = monitor_data[image_number] if monitor_data else None
                        scan_sum.add_to_sum(image, monitor)
                        full_sum.add_to_sum(image, monitor)

                        if save_scan_sums and scan_sum.nb_images == block_size:
                            dset = save_sum(
                                results,
                                name=f"Scan{scan_number}-Images{image_number - scan_sum.nb_images + 1}-{image_number}",
                                image_sum=scan_sum,
                            )
                            output_uris.append(f"{output_filename}::{dset.name}")

                            # Move to next sum
                            scan_sum.reset()

                    if save_scan_sums and scan_sum.nb_images > 0:
                        dset = save_sum(
                            results,
                            name=f"Scan{scan_number}-Images{image_number - scan_sum.nb_images + 1}-{image_number}",
                            image_sum=scan_sum,
                        )
                        output_uris.append(f"{output_filename}::{dset.name}")
                        # Move to next sum
                        scan_sum.reset()

                if save_full_sum:
                    dset = save_sum(
                        results,
                        name=f"Sum of scans {scan_range.start} to {scan_range.stop - scan_range.step}",
                        image_sum=full_sum,
                    )
                    output_uris.append(f"{output_filename}::{dset.name}")

        self.outputs.output_uris = output_uris



import os
import re
import json
from typing import Union, Dict, Optional, Any, Iterator, Tuple, List
from collections.abc import Mapping, Sequence

import numpy
import h5py
from pyFAI import version as pyfai_version
from ewoks import __version__ as ewoks_version
from pyFAI.io.ponifile import PoniFile
from pyFAI.units import Unit
from silx.io.dictdump import dicttonx
from ewokscore.task import TaskInputError
from ewoksdata.data import nexus
from ewokscore.missing_data import is_missing_data

from .xrpd_utils import energy_wavelength
from .data_utils import data_from_storage

_REPLACE_PATTERNS = {
    "/gpfs/[^/]+/": "/",
    "/mnt/multipath-shares/": "/",
    "/lbsram/": "/",
}


def parse_units(radial_units: Any) -> Tuple[str, str]:
    if isinstance(radial_units, Unit):
        radial_units = radial_units.name
    if isinstance(radial_units, numpy.ndarray):
        radial_units = radial_units.item()
    if not isinstance(radial_units, str):
        raise TypeError(type(radial_units))
    unit_tuple = tuple(radial_units.split("_"))
    if len(unit_tuple) != 2:
        raise ValueError(f"Expected unit to be of the form X_Y. Got {radial_units}")
    return unit_tuple


def read_config(
    filename: Optional[str], replace_patterns: Optional[Dict[str, str]] = None
) -> dict:
    if not filename:
        return dict()
    if filename.endswith(".json"):
        parameters = _read_json(filename)
    else:
        parameters = _read_poni(filename)
    return normalize_parameters(parameters, replace_patterns=replace_patterns)


def _read_json(filename: str) -> dict:
    with open(filename, "r") as fp:
        return json.load(fp)


def _read_poni(filename: str) -> dict:
    return PoniFile(filename).as_dict()


def normalize_parameters(
    parameters: Union[str, int, float, Mapping, Sequence],
    replace_patterns: Optional[Dict[str, str]] = None,
) -> Union[str, int, float, Mapping, Sequence]:
    if replace_patterns is None:
        replace_patterns = _REPLACE_PATTERNS
    if isinstance(parameters, str):
        for pattern, repl in replace_patterns.items():
            parameters = re.sub(pattern, repl, parameters)
        return parameters
    if isinstance(parameters, Mapping):
        return {
            k: normalize_parameters(v, replace_patterns=replace_patterns)
            for k, v in parameters.items()
        }
    if isinstance(parameters, Sequence):
        return [
            normalize_parameters(v, replace_patterns=replace_patterns)
            for v in parameters
        ]
    return parameters


def extract_mask(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the mask:

    * mask_file: URL or None
    * do_mask: use mask_file or not
    * mask: URL, array or None (overwrites mask_file and do_mask when not None)
    """
    # Typically from a JSON configuration file
    mask_file = integration_options.get("mask_file", None)
    do_mask = integration_options.get("do_mask", None)
    if not do_mask:
        mask_file = None

    # Typically in memory from a previous calculation
    mask = integration_options.pop("mask", None)
    if isinstance(mask, str):
        mask_file = mask
        mask = None
        do_mask = True

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["mask_file"] = mask_file
    integration_options["do_mask"] = do_mask
    return mask


def extract_flatfield(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the flatfield:

    * flat_field: URL(s) (list of strings or comma separate string) or None
    * do_flat: use flat_field or not
    * flatfield: URL(s) (list of strings or comma separate string), array or None (overwrites flat_field and do_flat when not None)

    When a numpy array is returned, the flat field is removed from the integration options.
    """
    # Typically from a JSON configuration file
    flat_field = integration_options.get("flat_field", None)
    do_flat = integration_options.get("do_flat", None)
    if not do_flat:
        flat_field = None
    if not flat_field:
        do_flat = False

    # Typically in memory from a previous calculation
    flatfield = integration_options.pop("flatfield", None)
    if isinstance(flatfield, str):
        flatfield = [s.strip() for s in flatfield.split(",")]
    if isinstance(flatfield, Sequence) and isinstance(flatfield[0], str):
        flat_field = flatfield
        flatfield = None
        do_flat = True
    if flatfield is not None:
        flat_field = None
        do_flat = False

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["flat_field"] = flat_field
    integration_options["do_flat"] = do_flat
    return flatfield


def extract_darkcurrent(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the dark-current:

    * dark_current: URL(s) (list of strings or comma separate string) or None
    * do_dark: use flat_field or not
    * darkcurrent: URL(s) (list of strings or comma separate string), array or None (overwrites dark_current and do_dark when not None)

    When a numpy array is returned, the dark current is removed from the integration options.
    """
    # Typically from a JSON configuration file
    dark_current = integration_options.get("dark_current", None)
    do_dark = integration_options.get("do_dark", None)
    if not do_dark:
        dark_current = None
    if not dark_current:
        do_dark = False

    # Typically in memory from a previous calculation
    darkcurrent = integration_options.pop("darkcurrent", None)
    if isinstance(darkcurrent, str):
        darkcurrent = [s.strip() for s in darkcurrent.split(",")]
    if isinstance(darkcurrent, Sequence) and isinstance(darkcurrent[0], str):
        dark_current = darkcurrent
        darkcurrent = None
        do_dark = True
    if darkcurrent is not None:
        dark_current = None
        do_dark = False

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["dark_current"] = dark_current
    integration_options["do_dark"] = do_dark
    return darkcurrent


def compile_integration_info(parameters: Mapping, **extra) -> Dict[str, Any]:
    """Compile information on a pyFAI integration process. Add and rename keys when appropriate."""
    integration_info = dict(parameters)

    mask = extract_mask(integration_info)
    flatfield = extract_flatfield(integration_info)
    darkcurrent = extract_darkcurrent(integration_info)

    # Do not save the in-memory make, flat field and dark current
    if mask is not None:
        integration_info["mask_file"] = "[...]"
    if flatfield is not None:
        integration_info["flat_field"] = "[...]"
    if darkcurrent is not None:
        integration_info["dark_current"] = "[...]"

    for k, v in extra.items():
        if v is not None:
            integration_info[k] = v
    wavelength = integration_info.get("wavelength")
    if wavelength is not None:
        integration_info["energy"] = energy_wavelength(wavelength)
    return integration_info


def integration_info_as_text(integration_info: Mapping, **extra) -> List[str]:
    """Convert to a flat list of strings with the format `{key} = {value}`.
    Add keys and units when appropriate.
    """
    flatdict = {"pyfai_version": pyfai_version, "ewoks_version": ewoks_version}
    flatdict.update(integration_info)
    _add_extra(flatdict, extra)
    flatdict = dict(_flatten_dict(flatdict))

    energy = flatdict.pop("energy", None)
    if energy:
        flatdict["energy"] = f"{energy:.18e} keV"

    wavelength = flatdict.pop("wavelength", None)
    if wavelength is not None:
        flatdict["wavelength"] = f"{wavelength:.18e} m"

    geometry_dist = flatdict.pop("geometry.dist", None)
    if geometry_dist is not None:
        flatdict["distance"] = f"{geometry_dist:.18e} m"

    geometry_poni1 = flatdict.pop("geometry.poni1", None)
    if geometry_poni1 is not None:
        flatdict["center dim0"] = f"{geometry_poni1:.18e} m"

    geometry_poni2 = flatdict.pop("geometry.poni2", None)
    if geometry_poni2 is not None:
        flatdict["center dim1"] = f"{geometry_poni2:.18e} m"

    geometry_rot1 = flatdict.pop("geometry.rot1", None)
    if geometry_rot1 is not None:
        flatdict["rot1"] = f"{geometry_rot1:.18e} rad"

    geometry_rot2 = flatdict.pop("geometry.rot2", None)
    if geometry_rot2 is not None:
        flatdict["rot2"] = f"{geometry_rot2:.18e} rad"

    geometry_rot3 = flatdict.pop("geometry.rot3", None)
    if geometry_rot3 is not None:
        flatdict["rot3"] = f"{geometry_rot3:.18e} rad"

    return [f"{k} = {v}" for k, v in flatdict.items()]


def integration_info_as_nxdict(
    integration_info, as_nxnote: bool = True
) -> Dict[str, Any]:
    """Convert to a Nexus dictionary. Add keys and units when appropriate."""
    configuration = {"ewoks_version": ewoks_version}
    configuration.update(integration_info)
    nxtree_dict = {
        "@NX_class": "NXprocess",
        "program": "pyFAI",
        "version": pyfai_version,
    }
    if as_nxnote:
        nxtree_dict["configuration"] = {
            "@NX_class": "NXnote",
            "type": "application/json",
            "data": json.dumps(configuration, cls=PyFaiEncoder),
        }
    else:
        configuration["@NX_class"] = "NXcollection"
        nxtree_dict["configuration"] = configuration
        if "energy" in configuration:
            configuration["energy@units"] = "keV"
        if "wavelength" in configuration:
            configuration["wavelength@units"] = "m"
        geometry = configuration.get("geometry", dict())
        if "dist" in geometry:
            geometry["dist@units"] = "m"
        if "poni1" in geometry:
            geometry["poni1@units"] = "m"
        if "poni2" in geometry:
            geometry["poni1@units"] = "m"
        if "rot1" in geometry:
            geometry["rot1@units"] = "rad"
        if "rot2" in geometry:
            geometry["rot2@units"] = "rad"
        if "rot3" in geometry:
            geometry["rot3@units"] = "rad"
    return nxtree_dict


class PyFaiEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (numpy.generic, numpy.ndarray)):
            return obj.tolist()
        return super().default(obj)


def _flatten_dict(
    adict: Mapping, _prefix: Optional[str] = None
) -> Iterator[Tuple[str, Any]]:
    if _prefix is None:
        _prefix = ""
    for k, v in adict.items():
        k = _prefix + k
        if isinstance(v, Mapping):
            yield from _flatten_dict(v, _prefix=f"{k}.")
        else:
            yield k, v


def _add_extra(adict: Mapping, extra: Mapping):
    for k, v in extra.items():
        if v is not None:
            adict[k] = v


def create_nxprocess(
    master_parent: h5py.Group, data_parent: h5py.Group, nxprocess_name: str, info
) -> h5py.Group:
    nxtree_dict = integration_info_as_nxdict(info)
    nxprocess_path = f"{data_parent.name}/{nxprocess_name}"
    dicttonx(nxtree_dict, data_parent.file, h5path=nxprocess_path, update_mode="modify")
    nxprocess = data_parent[nxprocess_name]
    if data_parent.file.filename != master_parent.file.filename:
        create_hdf5_link(nxprocess, master_parent, nxprocess_name)
    return nxprocess


def create_nxdata(
    nxprocess: h5py.Group,
    intensity_dim: int,
    radial,
    radial_units,
    azimuthal,
    nxdata_name: str = "integrated",
) -> h5py.Group:
    nxdata = nxprocess.create_group(nxdata_name)
    nxdata.attrs["NX_class"] = "NXdata"
    nxprocess.attrs["default"] = nxdata_name

    # Axes interpretation
    add_axes_to_nxdata(nxdata, intensity_dim, radial, radial_units, azimuthal)

    return nxdata


def add_axes_to_nxdata(
    nxdata: h5py.Group, intensity_dim: int, radial, radial_units, azimuthal
):
    # Axes names and units
    radial_units = data_from_storage(radial_units, remove_numpy=True)
    try:
        radial_name, radial_units = parse_units(radial_units)
    except ValueError as e:
        raise TaskInputError(e)
    has_azimuth = not is_missing_data(azimuthal) and azimuthal is not None
    azimuthal_name = "azimuth"
    azimuthal_units = "deg"

    if has_azimuth and intensity_dim == 2:
        nxdata.attrs["axes"] = [azimuthal_name, radial_name]
        nxdata.attrs["interpretation"] = "image"
    elif has_azimuth and intensity_dim == 3:
        nxdata.attrs["axes"] = [".", azimuthal_name, radial_name]
        nxdata.attrs["interpretation"] = "image"
    elif not has_azimuth and intensity_dim == 2:
        nxdata.attrs["axes"] = [".", radial_name]
        nxdata.attrs["interpretation"] = "spectrum"
    elif not has_azimuth and intensity_dim == 1:
        nxdata.attrs["axes"] = [radial_name]
        nxdata.attrs["interpretation"] = "spectrum"
    else:
        raise ValueError("Unrecognized data")

    dset = nxdata.create_dataset(radial_name, data=radial)
    dset.attrs["units"] = radial_units
    if has_azimuth:
        dset = nxdata.create_dataset(azimuthal_name, data=azimuthal)
        dset.attrs["units"] = azimuthal_units


def create_nxprocess_links(
    nxprocess: h5py.Group,
    target: h5py.Group,
    link_name: str,
    mark_as_default: bool = True,
) -> None:
    nxdata = nxprocess["integrated"]
    try:
        intensity = nxdata["intensity"]
    except KeyError:
        return
    create_hdf5_link(intensity, target, link_name)
    if mark_as_default:
        nexus.select_default_plot(intensity)
        if intensity.file.filename != nxprocess.file.filename:
            target.attrs["default"] = link_name
            nexus.select_default_plot(target)


def create_hdf5_link(
    source: Union[h5py.Dataset, h5py.Group], target: h5py.Group, link_name: str
) -> None:
    if link_name in target:
        return
    source_filename = source.file.filename
    target_filename = target.file.filename
    if source_filename == target_filename:
        target[link_name] = h5py.SoftLink(source.name)
    else:
        source_filename = os.path.relpath(
            source_filename, os.path.dirname(target_filename)
        )
        target[link_name] = h5py.ExternalLink(source_filename, source.name)



import re
import numpy
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksxrpd.ascii import OWSaveAsciiPattern1D


def test_save_ascii_task(tmpdir, setup1):
    assert_save_ascii(tmpdir, setup1, None)


def test_save_ascii_widget(tmpdir, setup1, qtapp):
    assert_save_ascii(tmpdir, setup1, qtapp)


def assert_save_ascii(tmpdir, setup1, qtapp):
    inputs = {
        "filename": str(tmpdir / "result.dat"),
        "x": numpy.linspace(1, 60, 60),
        "y": numpy.random.random(60),
        "xunits": "2th_deg",
        "header": {
            "energy": 10.2,
            "detector": setup1.detector,
            "geometry": setup1.geometry,
        },
        "metadata": {"name": "mysample"},
    }

    execute_task(
        OWSaveAsciiPattern1D.ewokstaskclass if qtapp is None else OWSaveAsciiPattern1D,
        inputs=inputs,
    )

    x, y = numpy.loadtxt(str(tmpdir / "result.dat")).T
    numpy.testing.assert_array_equal(x, inputs["x"])
    numpy.testing.assert_array_equal(y, inputs["y"])

    with open(tmpdir / "result.dat") as f:
        lines = list()
        for line in f:
            if not line.startswith("#"):
                break
            lines.append(line)
    lines = "".join(lines)

    for key in (
        "detector",
        "energy",
        "distance",
        "center dim0",
        "center dim1",
        "rot1",
        "rot2",
        "rot3",
        "xunits",
    ):
        assert f"{key} =" in lines
    assert "name = mysample" in lines
    m = re.findall("energy = (.+) keV", lines)
    assert len(m) == 1
    assert float(m[0]) == inputs["header"]["energy"]



from ewoksxrpd.gui import serialize
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_nexus
from ewoksxrpd.tasks.nexus import SaveNexusIntegrated


__all__ = ["OWSaveNexusIntegrated"]


class OWSaveNexusIntegrated(OWTriggerWidget, ewokstaskclass=SaveNexusIntegrated):
    name = "SaveNexusIntegrated"
    description = "Save result from integration in Nexus HDF5 format"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_nexus(self.get_default_input_values())
        parameter_info["radial"] = {
            "label": "Radial values",
            "value_for_type": "",
        }
        parameter_info["intensity"] = {
            "label": "Intensity values",
            "value_for_type": "",
        }
        parameter_info["radial_units"] = {
            "label": "Radial unit",
            "value_for_type": "",
            "serialize": str.lower,
        }
        parameter_info["info"] = {
            "label": "Configuration",
            "value_for_type": "",
            "serialize": serialize.json_dumps,
            "deserialize": serialize.json_loads,
        }
        parameter_info["azimuthal"] = {
            "label": "Azimuthal values",
            "value_for_type": "",
        }
        parameter_info["intensity_error"] = {
            "label": "Intensity errors",
            "value_for_type": "",
        }

        self._create_input_form(parameter_info)



from typing import Optional
import numpy

from ..tasks.integrate import Integrate1D
from .xrpd_theory import Measurement, Setup, xPattern, yPattern


def test_flatdark_task(
    imageSetup1SampleA: Measurement,
    setup1: Setup,
    xSampleA: xPattern,
    ySampleA: yPattern,
):
    image = imageSetup1SampleA.image
    assert_integrate1d(
        image,
        imageSetup1SampleA.monitor,
        setup1,
        xSampleA,
        ySampleA,
    )

    imax = image.max()
    flatfield = numpy.random.uniform(low=imax * 0.9, high=imax * 1.5, size=image.shape)
    darkcurrent = numpy.random.uniform(low=0, high=imax * 0.1, size=image.shape)

    # Flat/dark correction (pyFAI):
    #  Icor = (I - dark) / flat
    image = flatfield * image + darkcurrent
    assert_integrate1d(
        image,
        imageSetup1SampleA.monitor,
        setup1,
        xSampleA,
        ySampleA,
        flatfield=flatfield,
        darkcurrent=darkcurrent,
    )

    # Flat/dark correction (counts):
    #  Icor = (I - dark) / max(flat - dark, 1)
    assert_integrate1d(
        image,
        imageSetup1SampleA.monitor,
        setup1,
        xSampleA,
        ySampleA,
        flatfield=flatfield + darkcurrent,
        darkcurrent=darkcurrent,
        darkflatmethod="counts",
    )


def assert_integrate1d(
    image: numpy.ndarray,
    monitor: float,
    setup: Setup,
    xpattern: xPattern,
    ypattern: yPattern,
    flatfield: Optional[numpy.ndarray] = None,
    darkcurrent: Optional[numpy.ndarray] = None,
    **custom_integration_options
):
    integration_options = dict(xpattern.integration_options)
    integration_options.update(custom_integration_options)
    inputs = {
        "image": image,
        "detector": setup.detector,
        "geometry": setup.geometry,
        "energy": setup.energy,
        "monitor": monitor,
        "reference": ypattern.monitor,
        "integration_options": integration_options,
        "flatfield": flatfield,
        "darkcurrent": darkcurrent,
    }

    task = Integrate1D(inputs=inputs)
    task.execute()
    output_values = task.get_output_values()

    assert output_values["xunits"] == xpattern.units
    numpy.testing.assert_allclose(xpattern.x, output_values["x"], rtol=1e-6)
    atol = ypattern.y.max() * 0.01
    numpy.testing.assert_allclose(ypattern.y, output_values["y"], atol=atol)



from numbers import Number
import numpy
from numpy.typing import ArrayLike

from ewoksdata.data import bliss

from .morphology import binary_dilation
from .convolution import gaussian_filter
from .data_access import TaskWithDataAccess


__all__ = ["MaskDetection"]


def smooth_mask(mask: ArrayLike, hwhm: Number = 5) -> ArrayLike:
    """
    Extracted from `pyFAI.gui.cli_calibration`
    """
    big_mask = binary_dilation(mask, hwhm)
    sigma = hwhm / numpy.sqrt(2 * numpy.log(2))
    smooth_mask = gaussian_filter(big_mask.astype(numpy.float32), sigma)
    return smooth_mask.astype(bool)


class MaskDetection(
    TaskWithDataAccess,
    input_names=["image1", "monitor1", "image2", "monitor2"],
    optional_input_names=["smooth", "monitor_ratio_margin"],
    output_names=["mask"],
):
    """The pixels with the same ratio as the monitor ratio within an error margin
    are considered "good pixels". The others are masked off.

    The error margin `monitor_ratio_margin` is a fraction of the monitor ratio (0.1 by default).

    Masked pixels have value `1`. The mask can be smoothed with `smooth > 0` to avoid border effects.

    .. code:

        monitor_ratio = monitor_high / monitor_low
        image_ratio = image_high / image_low

        bad = abs(image_ratio - monitor_ratio) > monitor_ratio * monitor_ratio_margin
    """

    def run(self):
        monitor_high = self.get_data(self.inputs.monitor1)
        monitor_low = self.get_data(self.inputs.monitor2)
        image_high = bliss.get_image(self.inputs.image1)
        image_low = bliss.get_image(self.inputs.image2)
        if monitor_high < monitor_low:
            monitor_high, monitor_low = monitor_low, monitor_high
            image_high, image_low = image_low, image_high

        monitor_ratio = monitor_high / monitor_low
        with numpy.errstate(divide="ignore", invalid="ignore"):
            image_ratio = image_high / image_low
        monitor_ratio_margin = self.get_input_value("monitor_ratio_margin", default=0.1)
        threshold = monitor_ratio * monitor_ratio_margin
        mask = numpy.abs(image_ratio - monitor_ratio) > threshold
        mask |= ~numpy.isfinite(image_ratio)

        if self.inputs.smooth:
            mask = smooth_mask(mask, self.inputs.smooth)
        self.outputs.mask = mask.astype(numpy.int8)



from typing import Optional
import h5py
import numpy


class ImageSum:
    def __init__(self, shape) -> None:
        self.shape = shape
        self.summed_image = numpy.zeros(shape)
        self.summed_monitor = 0
        self.nb_images = 0

    def add_to_sum(self, image: numpy.ndarray, monitor: Optional[int]):
        self.summed_image += image
        if monitor is not None:
            self.summed_monitor += monitor
        self.nb_images += 1

    def reset(self):
        self.summed_image = numpy.zeros(self.shape)
        self.summed_monitor = 0
        self.nb_images = 0


def generate_range(start: int, end_arg: Optional[int], nitems: int) -> range:
    end = nitems if end_arg is None else end_arg + 1

    if (end - start) > nitems:
        raise ValueError(
            f"Asked range ({start},{end}) is bigger than number of items ({nitems})"
        )

    return range(start, end)


def save_sum(nxdata: h5py.Group, name: str, image_sum: ImageSum):
    dset = nxdata.create_dataset(name, data=image_sum.summed_image)
    dset.attrs["monitor"] = image_sum.summed_monitor
    dset.attrs["nb_images"] = image_sum.nb_images

    if "signal" not in nxdata:
        nxdata.attrs["signal"] = name

    return dset



import os
from typing import Dict, Iterable, Tuple, List
import pyFAI.units
import pyFAI.calibrant


def energy_wavelength(x: float) -> float:
    """keV to m and vice versa"""
    return pyFAI.units.hc * 1e-10 / x


GEOMETRY_PARAMETERS = {"dist", "poni1", "poni2", "rot1", "rot2", "rot3"}


def validate_geometry(geometry: dict):
    required = GEOMETRY_PARAMETERS
    existing = set(geometry.keys())
    missing = required - existing
    if missing:
        raise ValueError(f"geometry has missing parameters {sorted(missing)}")
    unexpected = existing - required
    if unexpected:
        raise ValueError(f"geometry has unexpected parameters {sorted(unexpected)}")


def calibrant_ring_labels(calibrant: pyFAI.calibrant.Calibrant) -> List[str]:
    path = calibrant.get_filename()
    try:
        # Introduced in pyfai v2023.08
        path = calibrant._get_abs_path(path)
    except AttributeError:
        pass
    if not path or not os.path.isfile(path):
        raise ValueError(f"No such calibrant file: {path}")

    labels = list()
    i = 0
    with open(path, "r") as fd:
        for line in fd:
            if line.startswith("#"):
                continue
            line = line.rstrip()
            label = line.rpartition("#")[-1]
            if not label or label in labels:
                label = f"ring{i}"
            label = label.strip()
            labels.append(label)
            i += 1
    return labels


def points_to_rings(
    points: Iterable[Tuple[float, float, int]], calibrant: pyFAI.calibrant.Calibrant
) -> Dict[int, Dict[str, list]]:
    rings = dict()
    labels = calibrant_ring_labels(calibrant)
    for p0, p1, i in points:
        i = int(i)
        try:
            label = labels[i]
        except IndexError:
            label = f"ring{i}"
        adict = rings.get(label)
        if adict:
            adict["p0"].append(p0)
            adict["p1"].append(p1)
        else:
            rings[label] = {"p0": [p0], "p1": [p1]}
    return rings



import json
import numpy
import h5py
import pytest

from silx.io.utils import h5py_read_dataset
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksxrpd.nexus_integrated import OWSaveNexusIntegrated


@pytest.fixture
def common_inputs(tmpdir, setup1, bliss_lab6_scan):
    bliss_scan_url = f"{bliss_lab6_scan}::/2.1"

    return {
        "url": str(tmpdir / "result.h5"),
        "radial": numpy.linspace(1, 60, 60),
        "intensity": numpy.random.random((10, 60)),
        "radial_units": "2th_deg",
        "info": {
            "energy": 10.2,
            "detector": setup1.detector,
            "geometry": setup1.geometry,
        },
        "metadata": {"dummy": {"test": "test"}},
        "bliss_scan_url": bliss_scan_url,
        "retry_timeout": 5,
    }


def test_save_nexus_task(common_inputs):
    assert_save_nexus_no_azimuth(common_inputs, None)


def test_save_nexus_task_with_azimuth(common_inputs):
    assert_save_nexus_with_azimuth(common_inputs, None)


def test_save_nexus_widget(common_inputs, qtapp):
    assert_save_nexus_no_azimuth(common_inputs, qtapp)


def test_save_nexus_widget_with_azimuth(common_inputs, qtapp):
    assert_save_nexus_with_azimuth(common_inputs, qtapp)


def assert_save_nexus_no_azimuth(common_inputs, qtapp):
    inputs = {
        "radial": numpy.linspace(1, 60, 60),
        "intensity": numpy.random.random((10, 60)),
        **common_inputs,
    }

    execute_task(
        OWSaveNexusIntegrated.ewokstaskclass
        if qtapp is None
        else OWSaveNexusIntegrated,
        inputs=inputs,
    )

    with h5py.File(inputs["url"]) as root:
        expected = {"instrument", "measurement", "integrate", "dummy"}
        nxprocess = root["results/integrate"]
        assert set(root["results"].keys()) == expected
        assert "azimuthal" not in nxprocess
        assert "intensity_errors" not in nxprocess
        numpy.testing.assert_array_equal(nxprocess["integrated/2th"], inputs["radial"])
        numpy.testing.assert_array_equal(
            nxprocess["integrated/intensity"], inputs["intensity"]
        )
        numpy.testing.assert_array_equal(
            root["results/measurement/integrated"], inputs["intensity"]
        )
        config = json.loads(nxprocess["configuration/data"][()])
        numpy.testing.assert_array_equal(config["energy"], inputs["info"]["energy"])
        assert h5py_read_dataset(root["results/dummy/test"]) == "test"


def assert_save_nexus_with_azimuth(common_inputs, qtapp):
    inputs = {
        "radial": numpy.linspace(1, 60, 60),
        "azimuthal": numpy.linspace(1, 100, 100),
        "intensity": numpy.random.random((10, 100, 60)),
        **common_inputs,
    }

    execute_task(
        OWSaveNexusIntegrated.ewokstaskclass
        if qtapp is None
        else OWSaveNexusIntegrated,
        inputs=inputs,
    )

    with h5py.File(inputs["url"]) as root:
        expected = {"instrument", "measurement", "integrate", "dummy"}
        nxprocess = root["results/integrate"]
        assert set(root["results"].keys()) == expected
        assert "intensity_errors" not in nxprocess
        numpy.testing.assert_array_equal(nxprocess["integrated/2th"], inputs["radial"])
        numpy.testing.assert_array_equal(
            nxprocess["integrated/azimuth"], inputs["azimuthal"]
        )
        numpy.testing.assert_array_equal(
            nxprocess["integrated/intensity"], inputs["intensity"]
        )
        numpy.testing.assert_array_equal(
            root["results/measurement/integrated"], inputs["intensity"]
        )
        config = json.loads(nxprocess["configuration/data"][()])
        numpy.testing.assert_array_equal(config["energy"], inputs["info"]["energy"])
        assert h5py_read_dataset(root["results/dummy/test"]) == "test"



import logging
from contextlib import contextmanager
from collections import OrderedDict
from typing import Iterable, Dict, Mapping, Tuple

import pyFAI
import pyFAI.worker
from pyFAI.average import average_images
from ewokscore.hashing import uhash

from .utils import pyfai_utils

_WORKER_POOL = None


logger = logging.getLogger(__name__)


class WorkerPool:
    """Pool with one worker per configuration up to a maximum number of workers."""

    def __init__(self, nworkers: int = 1) -> None:
        self._workers: Dict[int, pyFAI.worker.Worker] = OrderedDict()
        self.nworkers = nworkers

    @staticmethod
    def _worker_id(*args):
        return uhash(args)

    @property
    def nworkers(self):
        return self._nworkers

    @nworkers.setter
    def nworkers(self, value: int):
        self._nworkers = value
        self._check_pool_size()

    def _check_pool_size(self):
        while self._workers and len(self._workers) > self.nworkers:
            self._workers.popitem(last=False)

    @contextmanager
    def worker(
        self, options: Mapping, demo: bool = False
    ) -> Iterable[pyFAI.worker.Worker]:
        # TODO: deal with threads and subprocesses
        worker_options, integration_options = self._split_options(options)
        logger.info("Pyfai worker options: %s", worker_options)
        logger.info("Pyfai integration options: %s", integration_options)
        worker_id = self._worker_id(worker_options, integration_options, demo)
        worker = self._workers.pop(worker_id, None)
        if worker is None:
            logger.info("Creating a new pyfai worker")
            worker = self._create_worker(worker_options, integration_options, demo)
        self._workers[worker_id] = worker
        self._check_pool_size()
        logger.info("Pyfai integration method: %s", worker._method)
        yield worker

    def _split_options(self, options: Mapping) -> Tuple[dict, dict]:
        integration_options = dict(options)
        worker_keys = "integrator_name", "extra_options"
        worker_options = {
            k: integration_options.pop(k)
            for k in worker_keys
            if k in integration_options
        }
        nbpt_rad = integration_options.get("nbpt_rad", None)
        if nbpt_rad:
            nbpt_azim = integration_options.get("nbpt_azim", 1)
            worker_options.setdefault("shapeOut", (nbpt_azim, nbpt_rad))
        unit = integration_options.get("unit")
        if unit:
            worker_options["unit"] = unit
        return worker_options, integration_options

    @staticmethod
    def _create_worker(
        worker_options: Mapping, integration_options: Mapping, demo: bool
    ) -> pyFAI.worker.Worker:
        if demo:
            return DemoWorker(integration_options, worker_options)
        return EwoksWorker(integration_options, worker_options)


def _get_global_pool() -> WorkerPool:
    global _WORKER_POOL
    if _WORKER_POOL is None:
        _WORKER_POOL = WorkerPool()
    return _WORKER_POOL


def set_maximum_persistent_workers(nworkers: int) -> None:
    pool = _get_global_pool()
    pool.nworkers = nworkers


class EwoksWorker(pyFAI.worker.Worker):
    def __init__(self, integration_options: Mapping, worker_options: Mapping) -> None:
        super().__init__(**worker_options)
        self.output = "raw"
        self._i = 0

        integration_options = dict(integration_options)

        mask = pyfai_utils.extract_mask(integration_options)
        flatfield = pyfai_utils.extract_flatfield(integration_options)
        darkcurrent = pyfai_utils.extract_darkcurrent(integration_options)
        darkflatmethod = integration_options.pop("darkflatmethod", None)

        # Flat field correction:
        #   - default: Icor = (I - dark) / flat
        #   - counts:  Icor = (I - dark) / max(flat - dark, 1)
        if darkflatmethod is not None:
            if integration_options["flat_field"] and integration_options["do_flat"]:
                flatfield = average_images(
                    integration_options["flat_field"],
                    filter_="mean",
                    fformat=None,
                    threshold=0,
                )
                integration_options["flat_field"] = None
                integration_options["do_flat"] = False
            if integration_options["dark_current"] and integration_options["do_dark"]:
                darkcurrent = average_images(
                    integration_options["dark_current"],
                    filter_="mean",
                    fformat=None,
                    threshold=0,
                )
                integration_options["dark_current"] = None
                integration_options["do_dark"] = False
            if darkflatmethod == "counts":
                if flatfield is not None:
                    if darkcurrent is not None:
                        flatfield = flatfield - darkcurrent
                    flatfield[flatfield < 1] = 1

        provided = set(integration_options)
        self.set_config(integration_options, consume_keys=True)
        unused = {k: v for k, v in integration_options.items() if k in provided}
        if unused:
            logger.warning("Unused pyfai integration options: %s", unused)

        # Flat/dark correction:
        #   Icor = (I - darkcurrent) / flatfield
        if mask is not None:
            self.ai.detector.set_mask(mask)
        if flatfield is not None:
            self.ai.detector.set_flatfield(flatfield)
        if darkcurrent is not None:
            self.ai.detector.set_darkcurrent(darkcurrent)


class DemoWorker(EwoksWorker):
    def process(self, data, *args, **kwargs):
        return super().process(data[:-1, :-1], *args, **kwargs)


@contextmanager
def persistent_worker(
    integration_options: Mapping, demo: bool = False
) -> Iterable[pyFAI.worker.Worker]:
    """Get a worker for a particular configuration that stays in memory."""
    pool = _get_global_pool()
    with pool.worker(integration_options, demo) as worker:
        yield worker



import os
from AnyQt import QtWidgets
from AnyQt import QtGui

# from AnyQt.QtCore import Qt
from ewoksxrpd.tasks.diagnostics import DiagnoseCalibrateSingleResults
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_diagnose_singlecalib


__all__ = ["OWDiagnoseCalibrateSingleResults"]


class OWDiagnoseCalibrateSingleResults(
    OWTriggerWidget, ewokstaskclass=DiagnoseCalibrateSingleResults
):
    name = "DiagnoseCalibrateSingleResults"
    description = "Diagnose single-distance calibration"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_diagnose_singlecalib(
            self.get_default_input_values()
        )
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        layout = self._get_main_layout()
        self._label = QtWidgets.QLabel()
        layout.addWidget(self._label)
        super()._init_main_area()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._update_output_file()

    def _update_output_file(self):
        inputs = self.get_task_input_values()
        filename = inputs.get("filename")
        if not filename or not os.path.isfile(filename):
            self._label.clear()
            return

        pixmap = QtGui.QPixmap(filename)
        # pixmap = pixmap.scaled(10,5,Qt.KeepAspectRatio)
        self._label.setPixmap(pixmap)



from ewokscore.task import Task
from ewoksdata.data.bliss import last_lima_image


class LastLimaImage(
    Task, input_names=["db_name"], optional_input_names=["demo"], output_names=["image"]
):
    """Get the last Lima image from Redis"""

    def run(self):
        image = last_lima_image(self.inputs.db_name)
        if self.inputs.demo:
            image = image[:-1, :-1]
        self.outputs.image = image



from .data_access import TaskWithDataAccess

__all__ = ["SubtractBackground"]


class SubtractBackground(
    TaskWithDataAccess,
    input_names=["image", "monitor", "background", "background_monitor"],
    output_names=["image", "monitor"],
):
    """Background removal with normalization

    .. code:

        Icor = I  - B / Bmon * Imon
    """

    def run(self):
        monitor = self.get_data(self.inputs.monitor)
        norm = monitor / self.get_data(self.inputs.background_monitor)
        background = norm * self.get_image(self.inputs.background)
        self.outputs.image = self.get_image(self.inputs.image) - background
        self.outputs.monitor = monitor



from typing import Dict, Mapping, Tuple
from silx.gui.plot import Plot2D
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_integrateblissscan
from ewoksxrpd.gui.forms import pack_geometry
from ewoksxrpd.gui.forms import unpack_geometry
from ewoksxrpd.gui.forms import unpack_enabled_geometry
from ewoksxrpd.tasks.integrate import IntegrateBlissScanWithoutSaving
from ewoksxrpd.tasks.utils import data_utils


__all__ = ["OWIntegrateBlissScanNoSave"]


class OWIntegrateBlissScanNoSave(
    OWTriggerWidget, ewokstaskclass=IntegrateBlissScanWithoutSaving
):
    name = "IntegrateBlissScanWithoutSaving"
    description = "1D or 2D integrate data from one detector of a single Bliss scan"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_integrateblissscan(
            self.get_default_input_values(), saving=False
        )
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        super()._init_main_area()

        layout = self._get_main_layout()
        self._plot = Plot2D(parent=self.mainArea)
        layout.addWidget(self._plot)
        self._slider = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._slider)
        self._slider.valueChanged[int].connect(self._select_output_image)
        self._plot_data = None

    def _refresh_non_form_output_widgets(self):
        outputs = self.get_task_output_values()
        data = outputs.get("intensity")
        if not data_utils.is_data(data):
            self._plot_data = None
            return
        self._plot_data = data_utils.convert_to_3d(data)
        self._slider.setMaximum(max(len(self._plot_data) - 1, 0))
        self._select_output_image(self._slider.value())

    def _select_output_image(self, select: int):
        if self._plot_data is not None:
            self._plot.addImage(self._plot_data[select], legend="image")

    def _values_from_form(
        self, values: Mapping, checked: Dict[str, bool], output: bool = False
    ) -> Mapping:
        return pack_geometry(values, checked)

    def _values_to_form(
        self, values: Mapping, output: bool = False
    ) -> Tuple[Mapping, Dict[str, bool]]:
        return unpack_geometry(values)

    def _enabled_to_form(
        self, enabled: Dict[str, bool], output: bool = False
    ) -> Dict[str, bool]:
        return unpack_enabled_geometry(enabled)



from ewoksorange.tests.utils import execute_task
from ewoksxrpd.tasks.sum import SumBlissScanImages, SumImages
import pytest
import os.path


@pytest.mark.parametrize("monitor_name", ("mon", None))
@pytest.mark.parametrize("skip_dark", [False, True])
def test_sum_bliss_scan(tmpdir, bliss_perkinelmer_scan, monitor_name, skip_dark):
    inputs = {
        "filename": str(bliss_perkinelmer_scan),
        "detector_name": "perkinelmer",
        "output_filename": str(tmpdir / "output.h5"),
        "scan": 2,
        "monitor_name": monitor_name,
        "skip_dark": skip_dark,
    }

    outputs = execute_task(
        SumBlissScanImages,
        inputs=inputs,
    )

    assert os.path.isfile(str(tmpdir / "output.h5"))
    assert isinstance(outputs["output_uri"], str)


@pytest.mark.parametrize("monitor_name", ("mon", None))
def test_sum(tmpdir, bliss_perkinelmer_scan, monitor_name):
    inputs = {
        "filename": str(bliss_perkinelmer_scan),
        "detector_name": "perkinelmer",
        "output_filename": str(tmpdir / "output.h5"),
        "start_scan": 2,
        "end_image": 10,
        "end_scan": 2,
        "monitor_name": monitor_name,
    }

    outputs = execute_task(
        SumImages,
        inputs=inputs,
    )

    assert os.path.isfile(str(tmpdir / "output.h5"))
    assert len(outputs["output_uris"]) == 1


def test_sum_type_both(tmpdir, bliss_perkinelmer_scan):
    inputs = {
        "filename": str(bliss_perkinelmer_scan),
        "detector_name": "perkinelmer",
        "output_filename": str(tmpdir / "output.h5"),
        "start_scan": 2,
        "end_scan": 2,
        "sum_type": "both",
    }

    outputs = execute_task(
        SumImages,
        inputs=inputs,
    )

    assert os.path.isfile(str(tmpdir / "output.h5"))
    # Scan sum and full sum
    assert len(outputs["output_uris"]) == 2



from os import PathLike
from typing import List
import numpy
import pytest
import pyFAI.azimuthalIntegrator
from ewoksorange.tests.utils import execute_task

from .xrpd_theory import Calibration, Setup

from orangecontrib.ewoksxrpd.calibratesingle import OWCalibrateSingle
from orangecontrib.ewoksxrpd.calibratemulti import OWCalibrateMulti
from orangecontrib.ewoksxrpd.calculategeometry import OWCalculateGeometry
from orangecontrib.ewoksxrpd.diagnose_singlecalib import (
    OWDiagnoseCalibrateSingleResults,
)
from orangecontrib.ewoksxrpd.diagnose_multicalib import (
    OWDiagnoseCalibrateMultiResults,
)


@pytest.mark.parametrize("fixed", [[], ["energy"], ["rot2", "dist"]])
def test_calibrate_single_distance_task(
    fixed: List[str],
    imageSetup1Calibrant1: Calibration,
    setup1: Setup,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    tmpdir: PathLike,
):
    assert_calibrate_single_distance(
        fixed, imageSetup1Calibrant1, setup1, aiSetup1, tmpdir, None
    )


@pytest.mark.parametrize("fixed", [[], ["energy"], ["rot2", "dist"]])
def test_calibrate_single_distance_widget(
    fixed: List[str],
    imageSetup1Calibrant1: Calibration,
    setup1: Setup,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    tmpdir: PathLike,
    qtapp,
):
    assert_calibrate_single_distance(
        fixed, imageSetup1Calibrant1, setup1, aiSetup1, tmpdir, qtapp
    )


@pytest.mark.parametrize("fixed", [[], ["energy"], ["rot2", "dist"]])
def test_calibrate_multi_distance_task(
    fixed: List[str],
    imageSetup1Calibrant1: Calibration,
    setup1: Setup,
    imageSetup2Calibrant1: Calibration,
    setup2: Setup,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    aiSetup2: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    tmpdir: PathLike,
):
    assert_calibrate_multi_distance(
        fixed,
        imageSetup1Calibrant1,
        setup1,
        imageSetup2Calibrant1,
        setup2,
        aiSetup1,
        aiSetup2,
        tmpdir,
        None,
    )


@pytest.mark.parametrize("fixed", [[], ["energy"], ["rot2", "dist"]])
def test_calibrate_multi_distance_widget(
    fixed: List[str],
    imageSetup1Calibrant1: Calibration,
    setup1: Setup,
    imageSetup2Calibrant1: Calibration,
    setup2: Setup,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    aiSetup2: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    tmpdir: PathLike,
    qtapp,
):
    assert_calibrate_multi_distance(
        fixed,
        imageSetup1Calibrant1,
        setup1,
        imageSetup2Calibrant1,
        setup2,
        aiSetup1,
        aiSetup2,
        tmpdir,
        qtapp,
    )


def assert_calibrate_single_distance(
    fixed: List[str],
    imageSetup1Calibrant1: Calibration,
    setup1: Setup,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    tmpdir: PathLike,
    qtapp: None,
):
    fixed = []
    geometry0, energy0 = guess_fit_parameters(
        setup1.geometry, setup1.energy, aiSetup1, fixed=fixed
    )
    inputs = {
        "image": imageSetup1Calibrant1.image,
        "geometry": geometry0,
        "detector": setup1.detector,
        "energy": energy0,
        "calibrant": imageSetup1Calibrant1.calibrant,
        "fixed": fixed,
    }

    outputs_values = execute_task(
        OWCalibrateSingle.ewokstaskclass if qtapp is None else OWCalibrateSingle,
        inputs=inputs,
        timeout=3,
    )
    assert_calibration(
        False,
        fixed,
        outputs_values["geometry"],
        outputs_values["energy"],
        outputs_values["rings"],
        setup1,
        imageSetup1Calibrant1,
        aiSetup1,
        tmpdir,
    )


def assert_calibrate_multi_distance(
    fixed: List[str],
    imageSetup1Calibrant1: Calibration,
    setup1: Setup,
    imageSetup2Calibrant1: Calibration,
    setup2: Setup,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    aiSetup2: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    tmpdir: PathLike,
    qtapp,
):
    images = [imageSetup1Calibrant1.image, imageSetup2Calibrant1.image]
    positionunits_in_meter = 1e-2  # cm
    positions = [
        setup1.geometry["dist"] / positionunits_in_meter,
        setup2.geometry["dist"] / positionunits_in_meter,
    ]
    geometry0, energy0 = guess_fit_parameters(
        setup1.geometry, setup1.energy, aiSetup1, fixed=fixed
    )
    reference_position = positions[0]

    inputs = {
        "images": images,
        "positions": positions,
        "positionunits_in_meter": positionunits_in_meter,
        "reference_position": reference_position,
        "geometry": geometry0,
        "detector": setup1.detector,
        "energy": energy0,
        "calibrant": imageSetup1Calibrant1.calibrant,
        "fixed": fixed,
    }

    calibresults = execute_task(
        OWCalibrateMulti.ewokstaskclass if qtapp is None else OWCalibrateMulti,
        inputs=inputs,
    )

    assert_calibration(
        True,
        fixed,
        calibresults["geometry"],
        calibresults["energy"],
        calibresults["rings"]["0"],
        setup1,
        imageSetup1Calibrant1,
        aiSetup1,
        tmpdir,
    )

    geometry = dict(calibresults["geometry"])
    shift_in_meter = setup2.geometry["dist"] - setup1.geometry["dist"]
    geometry["dist"] += shift_in_meter
    assert_calibration(
        True,
        fixed,
        geometry,
        calibresults["energy"],
        calibresults["rings"]["1"],
        setup2,
        imageSetup2Calibrant1,
        aiSetup2,
        tmpdir,
    )

    inputs = {
        "parametrization": calibresults["parametrization"],
        "parameters": calibresults["parameters"],
        "position": positions[0],
    }

    calcresults = execute_task(
        OWCalculateGeometry.ewokstaskclass if qtapp is None else OWCalculateGeometry,
        inputs=inputs,
    )

    assert_geometry(
        True,
        fixed,
        calcresults["geometry"],
        calcresults["energy"],
        setup1,
        aiSetup1,
    )

    inputs["position"] = positions[1]
    calcresults = execute_task(
        OWCalculateGeometry.ewokstaskclass if qtapp is None else OWCalculateGeometry,
        inputs=inputs,
    )

    assert_geometry(
        True,
        fixed,
        calcresults["geometry"],
        calcresults["energy"],
        setup2,
        aiSetup2,
    )

    # Set show=True to visualize the calibration results
    filename = tmpdir / "test.png"
    inputs = {
        "images": images,
        "positions": positions,
        "detector": setup1.detector,
        "calibrant": imageSetup1Calibrant1.calibrant,
        "rings": calibresults["rings"],
        "parametrization": calibresults["parametrization"],
        "parameters": calibresults["parameters"],
        "show": False,
        "filename": str(filename),
    }
    execute_task(
        OWDiagnoseCalibrateMultiResults.ewokstaskclass
        if qtapp is None
        else OWDiagnoseCalibrateMultiResults,
        inputs=inputs,
    )
    assert filename.exists()


def assert_calibration(
    multi_distance: bool,
    fixed: List[str],
    geometry: dict,
    energy: float,
    rings: dict,
    setup: Setup,
    calibration: Calibration,
    ai: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    tmpdir: PathLike,
    qtapp=None,
):
    assert_geometry(multi_distance, fixed, geometry, energy, setup, ai)

    filename = tmpdir / "diagnose.png"
    inputs = {
        "image": calibration.image,
        "geometry": geometry,
        "detector": setup.detector,
        "energy": energy,
        "calibrant": calibration.calibrant,
        "rings": rings,
        "show": False,
        "filename": str(filename),
    }

    execute_task(
        OWDiagnoseCalibrateSingleResults.ewokstaskclass
        if qtapp is None
        else OWDiagnoseCalibrateSingleResults,
        inputs=inputs,
        timeout=10,
    )
    assert filename.exists()


def guess_fit_parameters(
    geometry: dict,
    energy: float,
    ai: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    fixed: List[str],
) -> dict:
    """Start fitting with parameters that are reasonably close to the real values"""

    def modify(pname, real_value):
        if pname in fixed:
            return real_value
        elif pname.startswith("rot"):
            # wrong by 1 degree
            return real_value + numpy.radians(0.1)
        elif pname == "dist":
            # wrong by 1 mm
            return real_value + 1e-3
        elif pname == "poni1":
            # wrong by 5 pixels
            return real_value + 5 * ai.detector.pixel1
        elif pname == "poni2":
            # wrong by 5 pixels
            return real_value - 5 * ai.detector.pixel2
        elif pname == "energy":
            # wrong by 50 eV
            return real_value + 50e-3
        else:
            raise ValueError(pname)

    geometry = {
        pname: modify(pname, real_value) for pname, real_value in geometry.items()
    }
    return geometry, modify("energy", energy)


def assert_geometry(
    multi_distance: bool,
    fixed: List[str],
    geometry: dict,
    energy: float,
    setup: Setup,
    ai: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
):
    """Energy and distance cannot be distinguished by a single distance calibration"""
    expected_parameters = dict(setup.geometry)
    expected_parameters["energy"] = setup.energy
    parameters = dict(geometry)
    parameters["energy"] = energy
    assert parameters.keys() == expected_parameters.keys()
    for pname, expected in expected_parameters.items():
        if pname in fixed:
            margin = 0
        elif pname.startswith("rot"):
            # wrong by +/- 0.2 degree
            margin = numpy.radians(0.2)
        elif pname == "dist":
            if not multi_distance:
                continue
            # wrong by +/- 50 um
            margin = 50e-6
        elif pname == "energy":
            if not multi_distance:
                continue
            # wrong by +/- 5 eV
            margin = 5e-3
        elif pname == "poni1":
            # wrong by +/- 0.5 pixels
            margin = 0.5 * ai.detector.pixel1
        elif pname == "poni2":
            # wrong by +/- 0.5 pixels
            margin = 0.5 * ai.detector.pixel2
        else:
            raise ValueError(pname)
        assert pytest.approx(expected, abs=margin) == parameters[pname], pname



import os
from AnyQt import QtWidgets
from AnyQt import QtGui

from ewoksxrpd.tasks.diagnostics import DiagnoseIntegrate1D
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_diagnose_integrate1d


__all__ = ["OWDiagnoseIntegrate1D"]


class OWDiagnoseIntegrate1D(OWTriggerWidget, ewokstaskclass=DiagnoseIntegrate1D):
    name = "DiagnoseIntegrate1D"
    description = "Diagnose 1D diffraction pattern"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_diagnose_integrate1d(
            self.get_default_input_values()
        )
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        layout = self._get_main_layout()
        self._label = QtWidgets.QLabel()
        layout.addWidget(self._label)
        super()._init_main_area()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._update_output_file()

    def _update_output_file(self):
        inputs = self.get_task_input_values()
        filename = inputs.get("filename")
        if not filename or not os.path.isfile(filename):
            self._label.clear()
            return

        pixmap = QtGui.QPixmap(filename)
        self._label.setPixmap(pixmap)



from typing import Dict, List, Mapping, Optional, Tuple
from AnyQt import QtWidgets
from silx.gui.plot import ScatterView, PlotWidget
from ewoksdata.data import bliss

from ewoksxrpd.tasks.calibrate import CalibrateSingle
from ewoksxrpd.tasks.utils import data_utils
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui import plots
from ewoksxrpd.gui.forms import input_parameters_calibratesingle
from ewoksxrpd.gui.forms import output_parameters_calibratesingle
from ewoksxrpd.gui.forms import pack_geometry
from ewoksxrpd.gui.forms import unpack_geometry
from ewoksxrpd.gui.forms import unpack_enabled_geometry


__all__ = ["OWCalibrateSingle"]


class OWCalibrateSingle(OWTriggerWidget, ewokstaskclass=CalibrateSingle):
    name = "CalibrateSingle"
    description = "Single distance calibration"
    icon = "icons/widget.png"
    want_main_area = True

    def __init__(self, *args, **kwargs) -> None:
        self._tabs = QtWidgets.QTabWidget()
        self._show = {
            "input_rings": True,
            "output_rings": False,
            "detected_rings": False,
        }
        self._legends = dict()
        self._cache = dict()
        super().__init__(*args, **kwargs)

    def _init_forms(self):
        parameter_info = input_parameters_calibratesingle(
            self.get_default_input_values()
        )
        self._create_input_form(parameter_info)
        parameter_info = output_parameters_calibratesingle()
        self._create_output_form(parameter_info)

    def _init_control_area(self):
        super()._init_control_area()
        layout = self._get_control_layout()

        w = QtWidgets.QPushButton("Accept refined")
        layout.addWidget(w)
        w.released.connect(self._accept_refined_parameters)

        self._show_input_rings_widget = w = QtWidgets.QCheckBox("Guess rings")
        w.setChecked(self._show["input_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_input_rings)

        self._show_output_rings_widget = w = QtWidgets.QCheckBox("Refined rings")
        w.setChecked(self._show["output_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_output_rings)

        self._show_detected_rings_widget = w = QtWidgets.QCheckBox("Detected rings")
        w.setChecked(self._show["detected_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_detected_rings)

    def _set_show_input_rings(self):
        self._show["input_rings"] = self._show_input_rings_widget.isChecked()
        self._refresh_non_form_input_widgets()

    def _set_show_output_rings(self):
        self._show["output_rings"] = self._show_output_rings_widget.isChecked()
        self._refresh_non_form_output_widgets()

    def _set_show_detected_rings(self):
        self._show["detected_rings"] = self._show_detected_rings_widget.isChecked()
        self._refresh_non_form_output_widgets()

    def _init_main_area(self):
        layout = self._get_main_layout()
        plot = ScatterView()
        w = plot.getPlotWidget()
        w.setGraphXLabel("Dim 2 (pixels)")
        w.setGraphYLabel("Dim 1 (pixels)")
        self._tabs.addTab(plot, "Image")
        layout.addWidget(self._tabs)
        super()._init_main_area()

    def _add_output_form_widget(self):
        self._tabs.addTab(self._output_form, "Refined Geometry")

    def _refresh_non_form_input_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_input_widgets()
            self._refresh_input_plots()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._refresh_output_plots()

    def _accept_refined_parameters(self):
        energy = self.get_task_output_value("energy")
        if energy:
            self.update_default_inputs(energy=energy)
        geometry = self.get_task_output_value("geometry")
        if geometry:
            self.update_default_inputs(geometry=dict(geometry))
        self._refresh_input_widgets()

    def _input_form_edited(self):
        super()._input_form_edited()
        fixed = [
            k for k, v in self._input_form.get_parameters_checked().items() if not v
        ]
        self.update_default_inputs(fixed=fixed)
        self._refresh_non_form_input_widgets()

    def _refresh_input_plots(self):
        if self.plot is None:
            return None
        inputs = self.get_task_input_values()
        self._update_image(inputs)
        self._update_input_rings(inputs)

    def _refresh_output_plots(self):
        if self.plot is None:
            return None
        outputs = self.get_task_output_values()
        self._update_detected_rings(outputs)
        inoutputs = {**self.get_task_input_values(), **outputs}
        self._update_output_rings(inoutputs)

    def _update_image(self, inputs):
        image = inputs.get("image")
        if isinstance(image, str):
            previous_image_url = self._cache.get("image")
            if previous_image_url == image:
                return
            self._cache["image"] = image
        self._remove_from_plot("image")
        if not data_utils.is_data(image):
            return
        image = bliss.get_image(image, retry_timeout=0)
        self._legends["image"] = [plots.plot_image(self.plot, image, legend="image")]

    def _update_input_rings(self, values):
        self._remove_from_plot("input_rings")
        if not self._show["input_rings"]:
            return
        self._legends["input_rings"] = self._update_theoretical_rings(values, "input")

    def _update_output_rings(self, values):
        self._remove_from_plot("output_rings")
        if not self._show["output_rings"]:
            return
        self._legends["output_rings"] = self._update_theoretical_rings(values, "output")

    def _update_theoretical_rings(self, values, legend) -> List[str]:
        energy = values.get("energy")
        geometry = values.get("geometry")
        detector = values.get("detector")
        detector_config = values.get("detector_config")
        calibrant = values.get("calibrant")
        max_rings = values.get("max_rings")
        if not energy or not geometry or not detector or not calibrant:
            return list()
        geometry = data_utils.data_from_storage(geometry)
        return plots.plot_theoretical_rings(
            self.plot,
            detector,
            calibrant,
            energy,
            geometry,
            detector_config=detector_config,
            max_rings=max_rings,
            legend=legend,
        )

    def _update_detected_rings(self, values):
        self._remove_from_plot("detected_rings")
        if not self._show["detected_rings"]:
            return
        rings = values.get("rings")
        if not rings:
            return
        rings = data_utils.data_from_storage(rings)
        self._legends["detected_rings"] = plots.plot_detected_rings(self.plot, rings)

    def _remove_from_plot(self, name: str) -> None:
        legends = self._legends.pop(name, list())
        for legend in legends:
            self.plot.remove(legend=legend)

    def _values_from_form(
        self, values: Mapping, checked: Dict[str, bool], output: bool = False
    ) -> Mapping:
        return pack_geometry(values, checked)

    def _values_to_form(
        self, values: Mapping, output: bool = False
    ) -> Tuple[Mapping, Dict[str, bool]]:
        return unpack_geometry(values)

    def _enabled_to_form(
        self, enabled: Dict[str, bool], output: bool = False
    ) -> Dict[str, bool]:
        return unpack_enabled_geometry(enabled)

    @property
    def plot(self) -> Optional[PlotWidget]:
        if self._tabs.count() == 0:
            return None
        return self._tabs.widget(0).getPlotWidget()



from contextlib import contextmanager

import h5py
from silx.io.dictdump import dicttonx

from .utils import data_utils, pyfai_utils
from .data_access import TaskWithDataAccess

__all__ = ["SaveNexusPattern1D", "SaveNexusIntegrated"]


class _BaseSaveNexusIntegrated(
    TaskWithDataAccess,
    input_names=["url"],
    optional_input_names=[
        "bliss_scan_url",
        "metadata",
        "nxprocess_name",
        "nxmeasurement_name",
        "nxprocess_as_default",
        "external_url",
    ],
    output_names=["saved"],
    register=False,
):
    @property
    def _process_info(self):
        raise NotImplementedError

    @property
    def _nxprocess_name(self):
        if self.inputs.nxprocess_name:
            return self.inputs.nxprocess_name
        return "integrate"

    @property
    def _nxmeasurement_name(self):
        if self.inputs.nxmeasurement_name:
            return self.inputs.nxmeasurement_name
        return "integrated"

    @contextmanager
    def _save_context(self):
        master_url = self.inputs.url
        data_url = self.get_input_value("external_url", master_url)

        with self.open_h5item(data_url, mode="a", create=True) as data_parent:
            assert isinstance(data_parent, h5py.Group)
            with self.open_h5item(master_url, mode="a", create=True) as master_parent:
                assert isinstance(master_parent, h5py.Group)
                nxprocess = pyfai_utils.create_nxprocess(
                    master_parent, data_parent, self._nxprocess_name, self._process_info
                )

            yield nxprocess

            with self.open_h5item(master_url, mode="a", create=True) as master_parent:
                url = data_utils.data_from_storage(
                    self.inputs.bliss_scan_url, remove_numpy=True
                )
                if url:
                    self.link_bliss_scan(master_parent, url)
                mark_as_default = self.get_input_value("nxprocess_as_default", True)
                measurement = master_parent.require_group("measurement")
                measurement.attrs["NX_class"] = "NXcollection"
                pyfai_utils.create_nxprocess_links(
                    nxprocess,
                    measurement,
                    self._nxmeasurement_name,
                    mark_as_default=mark_as_default,
                )
                if self.inputs.metadata:
                    dicttonx(
                        self.inputs.metadata,
                        master_parent,
                        update_mode="add",
                        add_nx_class=True,
                    )
        self.outputs.saved = True


class SaveNexusPattern1D(
    _BaseSaveNexusIntegrated,
    input_names=["x", "y", "xunits"],
    optional_input_names=["header", "yerror"],
):
    """Save single diffractogram in HDF5/NeXus format"""

    def run(self):
        with self._save_context() as nxprocess:
            nxdata = pyfai_utils.create_nxdata(
                nxprocess, self.inputs.y.ndim, self.inputs.x, self.inputs.xunits, None
            )
            nxdata.attrs["signal"] = "intensity"
            nxdata["intensity"] = self.inputs.y
            if not self.missing_inputs.yerror:
                nxdata["intensity_errors"] = self.inputs.yerror

    @property
    def _process_info(self):
        return self.inputs.header


class SaveNexusIntegrated(
    _BaseSaveNexusIntegrated,
    input_names=["radial", "intensity", "radial_units"],
    optional_input_names=["info", "azimuthal", "intensity_error"],
):
    """Save 1D or 2D integration diffraction patterns in HDF5/NeXus format"""

    def run(self):
        with self._save_context() as nxprocess:
            nxdata = pyfai_utils.create_nxdata(
                nxprocess,
                self.inputs.intensity.ndim,
                self.inputs.radial,
                self.inputs.radial_units,
                self.inputs.azimuthal,
            )
            nxdata.attrs["signal"] = "intensity"
            nxdata["intensity"] = self.inputs.intensity
            if not self.missing_inputs.intensity_error:
                nxdata["intensity_errors"] = self.inputs.intensity_error

    @property
    def _process_info(self):
        return self.inputs.info



from AnyQt import QtWidgets
from silx.gui.plot import Plot2D
from ewoksdata.data import bliss

from ewoksxrpd.tasks.background import SubtractBackground
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.tasks.utils import data_utils
from ewoksxrpd.gui.forms import input_parameters_background


__all__ = ["OWSubtractBackground"]


class OWSubtractBackground(OWTriggerWidget, ewokstaskclass=SubtractBackground):
    name = "SubtractBackground"
    description = "Subtract background in 2D"
    icon = "icons/widget.png"
    want_main_area = True

    def __init__(self, *args, **kwargs) -> None:
        self._tabs = QtWidgets.QTabWidget()
        super().__init__(*args, **kwargs)

    def _init_forms(self):
        parameter_info = input_parameters_background(self.get_default_input_values())
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        layout = self._get_main_layout()
        layout.addWidget(self._tabs)
        for name in ("Image", "Background", "Subtracted"):
            self._tabs.addTab(Plot2D(), name)
        super()._init_main_area()

    def _refresh_non_form_input_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_input_widgets()
            self._refresh_input_plots()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._refresh_output_plots()

    def _refresh_input_plots(self):
        if self._tabs.count() == 0:
            return
        inputs = self.get_task_input_values()
        self._update_image(inputs)
        self._update_background(inputs)

    def _refresh_output_plots(self):
        if self._tabs.count() == 0:
            return
        outputs = self.get_task_output_values()
        self._update_subtracted(outputs)

    def _update_image(self, inputs):
        plot = self._tabs.widget(0)
        plot.remove(kind="image")
        image = inputs.get("image")
        if data_utils.is_data(image):
            plot.addImage(bliss.get_image(image, retry_timeout=0))

    def _update_background(self, inputs):
        plot = self._tabs.widget(1)
        plot.remove(kind="image")
        image = inputs.get("background")
        if data_utils.is_data(image):
            plot.addImage(bliss.get_image(image, retry_timeout=0))

    def _update_subtracted(self, outputs):
        plot = self._tabs.widget(2)
        plot.remove(kind="image")
        image = outputs.get("image")
        if data_utils.is_data(image):
            plot.addImage(bliss.get_image(image, retry_timeout=0))



import numpy
from numpy.typing import ArrayLike
from numbers import Number

try:
    from pyFAI.ext import morphology

    def binary_dilation(mask: ArrayLike, hwhm: Number):
        fwhm = int(round(2.0 * hwhm))
        return morphology.binary_dilation(mask.astype(numpy.int8), fwhm)

except ImportError:
    from scipy.ndimage import morphology

    def binary_dilation(mask: ArrayLike, hwhm: Number):
        fwhm = int(round(2.0 * hwhm))
        my, mx = numpy.ogrid[-fwhm : fwhm + 1, -fwhm : fwhm + 1]
        grow = (mx * mx + my * my) <= (4.0 * hwhm * hwhm)
        return morphology.binary_dilation(mask, grow)



import os
import h5py
import pytest
import numpy

import pyFAI.azimuthalIntegrator
import pyFAI.detectors
from ewoksorange.canvas.handler import OrangeCanvasHandler
from ewoksorange.tests.conftest import qtapp  # noqa F401
from ewoksdata.tests.data import save_bliss_scan

from . import xrpd_theory


@pytest.fixture(scope="session")
def setup1() -> xrpd_theory.Setup:
    # Detector size is approx. 0.18 x 0.18 m
    geometry = {
        "dist": 10e-2,  # 10 cm
        "poni1": 10e-2,  # 10 cm
        "poni2": 10e-2,  # 10 cm
        "rot1": numpy.radians(10),  # 10 deg
        "rot2": 0,  # 0 deg
        "rot3": 0,  # 0 deg
    }
    return xrpd_theory.Setup(detector="Pilatus1M", energy=12, geometry=geometry)


@pytest.fixture(scope="session")
def setup2(setup1) -> xrpd_theory.Setup:
    # setup1 with detector shifted 5 cm backwards
    geometry = dict(setup1.geometry)
    geometry["dist"] += 5e-2
    assert geometry["dist"] > 0
    return xrpd_theory.Setup(
        detector=setup1.detector, energy=setup1.energy, geometry=geometry
    )


@pytest.fixture(scope="session")
def aiSetup1(
    setup1: xrpd_theory.Setup,
) -> pyFAI.azimuthalIntegrator.AzimuthalIntegrator:
    detector = pyFAI.detectors.detector_factory(setup1.detector)
    return pyFAI.azimuthalIntegrator.AzimuthalIntegrator(
        detector=detector, **setup1.geometry
    )


@pytest.fixture(scope="session")
def aiSetup2(
    setup2: xrpd_theory.Setup,
) -> pyFAI.azimuthalIntegrator.AzimuthalIntegrator:
    detector = pyFAI.detectors.detector_factory(setup2.detector)
    return pyFAI.azimuthalIntegrator.AzimuthalIntegrator(
        detector=detector, **setup2.geometry
    )


def linspace(low_limit, high_limit, n):
    half_bin = (high_limit - low_limit) / (2 * n)
    return numpy.linspace(low_limit + half_bin, high_limit - half_bin, n)


@pytest.fixture(scope="session")
def xSampleA() -> xrpd_theory.xPattern:
    x = linspace(0, 30, 1024)
    return xrpd_theory.xPattern(x=x, low_limit=0, high_limit=30, units="2th_deg")


@pytest.fixture(scope="session")
def ySampleA(xSampleA: xrpd_theory.xPattern) -> xrpd_theory.yPattern:
    x = xSampleA.x
    y = numpy.zeros(x.size)
    peaks = list()
    s = 0.5
    for ufrac in [0.1, 0.4, 0.7]:
        A = 100
        u = x[0] + ufrac * (x[-1] - x[0])
        y += (
            A
            * numpy.exp(-((x - u) ** 2) / (2 * s**2))
            / (s * numpy.sqrt(2 * numpy.pi))
        )
        peaks.append((A, u, s))
    return xrpd_theory.yPattern(y=y, monitor=1000, theory=peaks)


@pytest.fixture(scope="session")
def xSampleB() -> xrpd_theory.xPattern:
    x = linspace(0, 180, 1024)
    return xrpd_theory.xPattern(x=x, low_limit=0, high_limit=180, units="2th_deg")


@pytest.fixture(scope="session")
def ySampleB(xSampleB: xrpd_theory.xPattern) -> xrpd_theory.yPattern:
    y = numpy.full(xSampleB.x.size, 10)
    return xrpd_theory.yPattern(y=y, monitor=1000, theory=10)


@pytest.fixture(scope="session")
def imageSetup1SampleA(
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    xSampleA: xrpd_theory.xPattern,
    ySampleA: xrpd_theory.yPattern,
) -> xrpd_theory.Measurement:
    return xrpd_theory.measurement(aiSetup1, xSampleA, ySampleA, mult=2)


@pytest.fixture(scope="session")
def imageSetup2SampleA(
    aiSetup2: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    xSampleA: xrpd_theory.xPattern,
    ySampleA: xrpd_theory.yPattern,
) -> xrpd_theory.Measurement:
    return xrpd_theory.measurement(aiSetup2, xSampleA, ySampleA, mult=2)


@pytest.fixture(scope="session")
def image1Setup1SampleB(
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    xSampleB: xrpd_theory.xPattern,
    ySampleB: xrpd_theory.yPattern,
) -> xrpd_theory.Measurement:
    return xrpd_theory.measurement(aiSetup1, xSampleB, ySampleB, mult=2)


@pytest.fixture(scope="session")
def image2Setup1SampleB(
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    xSampleB: xrpd_theory.xPattern,
    ySampleB: xrpd_theory.yPattern,
) -> xrpd_theory.Measurement:
    return xrpd_theory.measurement(aiSetup1, xSampleB, ySampleB, mult=3)


@pytest.fixture(scope="session")
def imageSetup1Calibrant1(
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    setup1: xrpd_theory.Setup,
) -> xrpd_theory.Calibration:
    return xrpd_theory.calibration("LaB6", aiSetup1, setup1)


@pytest.fixture(scope="session")
def imageSetup2Calibrant1(
    aiSetup2: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    setup2: xrpd_theory.Setup,
) -> xrpd_theory.Calibration:
    return xrpd_theory.calibration("LaB6", aiSetup2, setup2)


def next_scan_number(filename) -> int:
    if not os.path.exists(filename):
        return 1
    with h5py.File(filename, "r") as h5file:
        return int(max(map(float, h5file.keys()))) + 1


@pytest.fixture(scope="session")
def ewoks_orange_canvas(qtapp):  # noqa F811
    with OrangeCanvasHandler() as handler:
        yield handler


@pytest.fixture(scope="session")
def bliss_lab6_scan(tmpdir_factory, imageSetup1Calibrant1):
    tmpdir = tmpdir_factory.mktemp("sample_dataset")
    npoints_per_file = 3
    npoints = 31
    scannr = 2
    image = imageSetup1Calibrant1.image
    return save_bliss_scan(
        tmpdir,
        image,
        npoints_per_file,
        npoints,
        scannr,
        subscannr=1,
        lima_names=("p3",),
        counter_names=("diode1", "diode2"),
        sequence="multiply",
    )


@pytest.fixture(scope="session")
def bliss_perkinelmer_scan(tmpdir_factory, imageSetup1Calibrant1):
    tmpdir = tmpdir_factory.mktemp("sample_dataset")
    npoints_per_file = 3
    npoints = 31
    scannr = 2
    image = imageSetup1Calibrant1.image
    return save_bliss_scan(
        tmpdir,
        image,
        npoints_per_file,
        npoints,
        scannr,
        subscannr=1,
        lima_names=("perkinelmer",),
        counter_names=("mon",),
        sequence="multiply",
    )


@pytest.fixture(scope="session")
def bliss_task_inputs(bliss_lab6_scan, setup1) -> dict:
    return {
        "filename": str(bliss_lab6_scan),
        "scan": 2,
        "detector": setup1.detector,
        "energy": setup1.energy,
        "geometry": setup1.geometry,
        "detector_name": "p3",
        "monitor_name": "monitor",
        "reference": 1.0,
        "retry_timeout": 2,
    }



import logging
from numbers import Number
from typing import Dict, List, Optional, Set, Tuple

import pyFAI.detectors
import pyFAI.calibrant
import pyFAI.goniometer

from .utils import xrpd_utils
from .utils import data_utils
from .data_access import TaskWithDataAccess


__all__ = ["CalibrateSingle", "CalibrateMulti", "CalculateGeometry"]


logger = logging.getLogger(__name__)


def parse_fixed(
    fixed: Optional[List[str]], parametrization: Dict[str, List[str]]
) -> Set[str]:
    if not fixed:
        return set()
    allowed = {"dist", "poni1", "poni2", "rot1", "rot2", "rot3", "energy"}
    existing = set(fixed)
    unexpected = existing - allowed
    if unexpected:
        raise ValueError(f"'fixed' has unexpected parameters {sorted(unexpected)}")
    out = set()
    for param in fixed:
        out |= set(parametrization.get(param, {param}))
    return out


class CalibrateSingle(
    TaskWithDataAccess,
    input_names=["image", "detector", "calibrant", "geometry", "energy"],
    optional_input_names=["detector_config", "fixed", "max_rings", "robust"],
    output_names=["geometry", "energy", "detector", "detector_config", "rings", "chi2"],
):
    """Single distance and energy calibration."""

    def run(self):
        detector = data_utils.data_from_storage(self.inputs.detector)
        detector_config = data_utils.data_from_storage(
            self.get_input_value("detector_config", None)
        )
        detector_object = pyFAI.detectors.detector_factory(
            detector, config=detector_config
        )
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
        geometry0 = data_utils.data_from_storage(self.inputs.geometry)
        xrpd_utils.validate_geometry(geometry0)
        wavelength0 = xrpd_utils.energy_wavelength(self.inputs.energy)

        # Define setup
        setup = pyFAI.goniometer.SingleGeometry(
            "single",
            image=self.get_image(self.inputs.image),
            detector=detector_object,
            geometry=dict(geometry0),
            calibrant=calibrant_object,
        )

        setup.set_wavelength(wavelength0)

        # Find diffraction rings and fit
        self._refine(setup)

        # Parse results
        results = setup.geometry_refinement.get_config()
        geometry = {
            k: v for k, v in results.items() if k in xrpd_utils.GEOMETRY_PARAMETERS
        }
        energy = xrpd_utils.energy_wavelength(results["wavelength"])
        rings = xrpd_utils.points_to_rings(
            setup.geometry_refinement.data, calibrant_object
        )

        self.outputs.geometry = geometry
        self.outputs.energy = energy
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.rings = rings
        if setup.geometry_refinement.data.ndim != 2:
            self.outputs.chi2 = float("nan")
        else:
            self.outputs.chi2 = setup.geometry_refinement.chi2()

    def _refine(self, setup: pyFAI.goniometer.GoniometerRefinement) -> None:
        """Find diffraction rings and fit"""
        parametrization = {"energy": ["wavelength"]}
        force_fixed = parse_fixed(self.inputs.fixed, parametrization)
        fixed = fixed_prev = set(xrpd_utils.GEOMETRY_PARAMETERS) | {
            "wavelength"
        }  # all fixed

        max_rings = self.inputs.max_rings
        if max_rings:
            if max_rings < 0:
                max_rings = None
        else:
            max_rings = None

        def fit(*release):
            nonlocal fixed, fixed_prev
            set_free = set(release) - force_fixed
            if not set_free:
                return
            fixed -= set_free
            setup.extract_cp(max_rings=max_rings)
            if setup.geometry_refinement.data.ndim != 2:
                logger.warning("No rings detected for calibration")
                return
            setup.geometry_refinement.refine3(fix=fixed)  # method="slsqp"
            fixed_prev = fixed

        if self.inputs.robust:
            releases = [
                ["dist"],
                ["poni1", "poni2"],
                ["rot1", "rot2", "rot3"],
                ["wavelength"],
            ]
        else:
            releases = [
                ["dist", "poni1", "poni2", "rot1", "rot2", "rot3", "wavelength"]
            ]
        n = len(releases)
        for i, release in enumerate(releases, 1):
            fit(*release)
            self.progress = i / n * 100


class CalibrateMulti(
    TaskWithDataAccess,
    input_names=[
        "images",
        "positions",
        "detector",
        "calibrant",
        "geometry",
        "energy",
    ],
    optional_input_names=[
        "reference_position",
        "sample_position",
        "positionunits_in_meter",
        "detector_config",
        "fixed",
        "max_rings",
        "robust",
    ],
    output_names=[
        "geometry",
        "energy",
        "detector",
        "detector_config",
        "rings",
        "chi2",
        "parametrization",
        "parameters",
    ],
):
    """Single energy, multi distance calibration with 1 distance motor.

    The `images` and `positions` are the diffraction patterns and the motor positions at which those were measured.

    The `reference_position` is the detector position that corresponds to `geometry` (`positions[0]` by default).

    The `sample_position` is the position where sample and detector would theoretically coincide (0 by default).

    The output `energy` and `geometry` correspond to `sample_position`.

    The units of `positions`, `sample_position` and `reference_position` are defined by `positionunits_in_meter`
    (1e-3 by default which means the unit is in millimeter).

    The parametrization allow deriving energy+geometry from the detector position.

    .. code-block:: python

        energy, geometry = parametrization(parameters, position)
    """

    def run(self):
        if len(self.inputs.images) != len(self.inputs.positions):
            raise ValueError("number of 'images' and 'positions' must be the same")
        detector = data_utils.data_from_storage(self.inputs.detector)
        detector_config = data_utils.data_from_storage(
            self.get_input_value("detector_config", None)
        )
        detector_object = pyFAI.detectors.detector_factory(
            detector, config=detector_config
        )
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
        energy0 = self.inputs.energy
        wavelength0 = xrpd_utils.energy_wavelength(energy0)
        geometry0 = data_utils.data_from_storage(self.inputs.geometry)
        xrpd_utils.validate_geometry(geometry0)

        # Parameterize the detector geometry
        trans_function, initial_guess = self._reparametrization(geometry0, energy0)

        # Define setup
        setup = pyFAI.goniometer.GoniometerRefinement(
            initial_guess,
            pos_function=lambda x: x,
            trans_function=trans_function,
            detector=detector_object,
            wavelength=wavelength0,
        )
        for i, (image, xi) in enumerate(zip(self.inputs.images, self.inputs.positions)):
            label = f"position{i}"
            image = self.get_image(image)
            xi = self.get_data(xi)
            setup.new_geometry(
                label, image=image, metadata=xi, calibrant=calibrant_object
            )

        # Find diffraction rings and fit
        self._refine(setup)

        # Parse results
        geometry, energy = self._get_fit_results(setup)
        paramerization = self._get_parameterization(setup)
        parameters = self._get_parameters(setup)
        rings = self._get_rings(setup, calibrant_object)
        chi2 = setup.chi2()

        self.outputs.geometry = geometry
        self.outputs.energy = energy
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.rings = rings
        self.outputs.chi2 = chi2
        self.outputs.parametrization = paramerization
        self.outputs.parameters = parameters

    def _reparametrization(
        self, geometry0: Dict, energy0: float
    ) -> Tuple[pyFAI.goniometer.ExtendedTransformation, List[Number]]:
        # xunits are the position units
        # dunits are the distance units of the fit parameters
        # aunits are the angular units of the fit parameters
        dunits_per_m = 100  # dunits/m
        aunits_per_rad = 10  # aunits/rad
        m_per_xunits = self._get_position_units()  # m/xunits

        trans_function = pyFAI.goniometer.ExtendedTransformation(
            param_names=[
                "dist_offset",  # dunits
                "dist_scale",  # dunits/xunits
                "poni1_offset",  # dunits
                "poni1_scale",  # dunits/xunits
                "poni2_offset",  # dunits
                "poni2_scale",  # dunits/xunits
                "arot1",  # aunits
                "arot2",  # aunits
                "arot3",  # aunits
                "energy",  # keV
            ],
            pos_names=["x"],  # xunits
            dist_expr=f"dist_scale / {dunits_per_m} * x + dist_offset / {dunits_per_m}",
            poni1_expr=f"poni1_scale / {dunits_per_m} * x + poni1_offset / {dunits_per_m}",
            poni2_expr=f"poni2_scale / {dunits_per_m} * x + poni2_offset / {dunits_per_m}",
            rot1_expr=f"arot1 / {aunits_per_rad}",
            rot2_expr=f"arot2 / {aunits_per_rad}",
            rot3_expr=f"arot3 / {aunits_per_rad}",
            wavelength_expr="hc*1e-10/energy",
        )

        xref = self._get_reference_position()  # xunits
        dist_offset = self._get_sample_position() * m_per_xunits  # m
        dist_scale = (geometry0["dist"] - dist_offset) / xref  # m/xunits

        poni1_scale = 0.0  # m/xunits
        poni1_offset = geometry0["poni1"]  # m
        poni2_scale = 0.0  # m/xunits
        poni2_offset = geometry0["poni2"]  # m

        initial_guess = [
            dist_offset * dunits_per_m,
            dist_scale * dunits_per_m,
            poni1_offset * dunits_per_m,
            poni1_scale * dunits_per_m,
            poni2_offset * dunits_per_m,
            poni2_scale * dunits_per_m,
            geometry0["rot1"] * aunits_per_rad,
            geometry0["rot2"] * aunits_per_rad,
            geometry0["rot3"] * aunits_per_rad,
            energy0,
        ]
        return trans_function, initial_guess

    def _refine(self, setup: pyFAI.goniometer.GoniometerRefinement) -> None:
        """Find diffraction rings and fit"""
        parametrization = {
            "dist": ["dist_offset", "dist_scale"],
            "poni1": ["poni1_offset", "poni1_scale"],
            "poni2": ["poni2_offset", "poni2_scale"],
            "rot1": ["arot1"],
            "rot2": ["arot2"],
            "rot3": ["arot3"],
        }
        force_fixed = parse_fixed(self.inputs.fixed, parametrization)
        fixed = fixed_prev = set(setup.trans_function.param_names)  # all fixed

        max_rings = self.inputs.max_rings
        if max_rings:
            if max_rings < 0:
                max_rings = None
        else:
            max_rings = None

        def fit(*release):
            nonlocal fixed, fixed_prev
            set_free = set(release) - force_fixed
            if not set_free:
                return
            fixed -= set_free
            for setup_single_dist in setup.single_geometries.values():
                setup_single_dist.extract_cp(max_rings=max_rings)
            setup.refine3(fix=fixed)  # method="slsqp"
            fixed_prev = fixed

        if self.inputs.robust:
            releases = [
                ["dist_offset", "dist_scale"],
                ["poni1_offset", "poni2_offset"],
                ["arot1", "arot2", "arot3"],
                ["poni1_scale", "poni2_scale"],
                ["energy"],
            ]
        else:
            releases = [
                [
                    "dist_offset",
                    "dist_scale",
                    "poni1_offset",
                    "poni2_offset",
                    "arot1",
                    "arot2",
                    "arot3",
                    "poni1_scale",
                    "poni2_scale",
                    "energy",
                ]
            ]
        n = len(releases)
        for i, release in enumerate(releases, 1):
            fit(*release)
            self.progress = i / n * 100

    def _get_rings(
        self,
        setup: pyFAI.goniometer.GoniometerRefinement,
        calibrant: pyFAI.calibrant.Calibrant,
    ) -> Dict[str, dict]:
        rings = dict()
        for i, setup_single_dist in enumerate(setup.single_geometries.values()):
            points = setup_single_dist.geometry_refinement.data
            rings[str(i)] = xrpd_utils.points_to_rings(points, calibrant)
        return rings

    def _get_fit_results(
        self, setup: pyFAI.goniometer.GoniometerRefinement
    ) -> Tuple[dict, Number]:
        parameters = setup.param
        xref = self._get_reference_position()
        results = setup.trans_function(parameters, xref)
        geometry = {k: getattr(results, k) for k in xrpd_utils.GEOMETRY_PARAMETERS}
        energy = xrpd_utils.energy_wavelength(results.wavelength)
        return geometry, energy

    def _get_parameterization(
        self, setup: pyFAI.goniometer.GoniometerRefinement
    ) -> dict:
        parametrization = dict(setup.trans_function.to_dict())
        parametrization.pop("constants")
        parametrization.pop("content")
        return parametrization

    def _get_parameters(self, setup: pyFAI.goniometer.GoniometerRefinement) -> dict:
        parameters = setup.param
        names = setup.trans_function.param_names
        return dict(zip(names, parameters))

    def _get_reference_position(self):
        """first image position by default (xunits)"""
        if self.missing_inputs.reference_position:
            return self.get_data(self.inputs.positions[0])
        else:
            return self.get_data(self.inputs.reference_position)

    def _get_sample_position(self):
        """0 by default (xunits)"""
        if self.missing_inputs.sample_position:
            return 0
        else:
            return self.get_data(self.inputs.sample_position)

    def _get_position_units(self):
        """millimeter by default"""
        if self.inputs.positionunits_in_meter:
            return self.inputs.positionunits_in_meter  # m/xunits
        else:
            return 1e-3  # m/xunits


def calculate_geometry(
    parametrization: dict, parameters: dict, position: Number
) -> Tuple[dict, Number]:
    trans_function = pyFAI.goniometer.ExtendedTransformation(**parametrization)
    parameters = [parameters[name] for name in trans_function.param_names]
    trans_function(parameters, position)
    results = trans_function(parameters, position)
    geometry = {k: getattr(results, k) for k in xrpd_utils.GEOMETRY_PARAMETERS}
    energy = xrpd_utils.energy_wavelength(results.wavelength)
    return geometry, energy


class CalculateGeometry(
    TaskWithDataAccess,
    input_names=["parametrization", "parameters", "position"],
    output_names=["geometry", "energy"],
):
    """Calculate energy and geometry from pyFAI parametrization"""

    def run(self):
        parametrization = data_utils.data_from_storage(self.inputs.parametrization)
        geometry, energy = calculate_geometry(
            parametrization,
            self.inputs.parameters,
            self.get_data(self.inputs.position),
        )
        self.outputs.geometry = geometry
        self.outputs.energy = energy
